import struct
import pytest
import moderngl


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in mat2 in_m;
            in vec2 in_v;

            out vec2 out_v;

            uniform float mult;

            void main() {
                out_v = in_m * in_v * mult;
            }
        ''',
        varyings=['out_v']
    )

    buf_m = ctx.buffer(struct.pack('4f', 1, 1, 1, 2))
    buf_v = ctx.buffer(struct.pack('2f', 4, 7))
    res = ctx.buffer(reserve=buf_v.size)

    vao = ctx.vertex_array(prog, [
        (buf_m, '4f', 'in_m'),
        (buf_v, '2f', 'in_v'),
    ])

    prog['mult'].value = 0.0
    vao.transform(res, moderngl.POINTS)
    x, y = struct.unpack('2f', res.read())
    assert pytest.approx(x) == 0.0
    assert pytest.approx(y) == 0.0

    prog['mult'].value = 1.0
    vao.transform(res, moderngl.POINTS)
    x, y = struct.unpack('2f', res.read())
    assert pytest.approx(x) == 11.0
    assert pytest.approx(y) == 18.0

    prog['mult'].value = 2.0
    vao.transform(res, moderngl.POINTS)
    x, y = struct.unpack('2f', res.read())
    assert pytest.approx(x) == 22.0
    assert pytest.approx(y) == 36.0



import pytest
import numpy as np
import moderngl as mgl


@pytest.fixture
def prog_render_depth_pass(ctx):
    return ctx.program(
        vertex_shader='''
                #version 330

                in vec3 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 1.0);
                }
            ''',
        fragment_shader='''
                #version 330

                void main() {
                }
            ''',
    )


@pytest.fixture
def vbo_triangle(ctx):
    vertices = np.array([[-1, 1, 0.], [-1, -1, 0.], [1, -1, 0.], ])
    return ctx.buffer(vertices.astype('f4').tobytes())


@pytest.fixture
def fbo_with_rasterised_triangle(ctx, vbo_triangle):
    def _build_fbo_with_rasterised_triangle(prog, size=(4, 4), depth_clear=1.0):
        tex_depth = ctx.depth_texture(size)  # implicit -> dtype='f4', components=1
        fbo_depth = ctx.framebuffer(depth_attachment=tex_depth)
        fbo_depth.use()
        fbo_depth.clear(depth=depth_clear)

        # vertex array object of triangle with depth pass prog
        vao_triangle = ctx.simple_vertex_array(prog, vbo_triangle, 'in_vert')
        # Now we render a triangle in there
        vao_triangle.render()

        return fbo_depth, tex_depth

    return _build_fbo_with_rasterised_triangle


@pytest.fixture
def np_triangle_rasterised():
    def _build_np_triangle_rastised(size):
        # It should have 0.5's where the triangle lies.
        depth_value_from_triangle_vertices = 0.0
        # Map [-1, +1] -> [0, 1]
        depth_value_in_depthbuffer = (depth_value_from_triangle_vertices + 1) * 0.5
        return np.array(
            [
                [depth_value_in_depthbuffer] * (size[0] - (j + 1)) + [1.0] * (j + 1)
                for j in range(size[1])
            ]
        )

    return _build_np_triangle_rastised


def test_cull_face(ctx,
                   prog_render_depth_pass,
                   vbo_triangle,
                   fbo_with_rasterised_triangle,
                   np_triangle_rasterised):

    size = (16,) * 2

    ctx.enable(mgl.DEPTH_TEST)
    ctx.enable(mgl.CULL_FACE)

    def _do_test_cull_face(cull_face: str, expected_results: np.array):
        ctx.cull_face = cull_face
        fbo_depth, tex_depth = fbo_with_rasterised_triangle(prog_render_depth_pass, size)
        depth_from_dbo = np.frombuffer(tex_depth.read(), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(depth_from_dbo, expected_results)

    ############################################################################
    # EXPECTED DATAS#
    # It should have 0.5's where the triangle lies.
    ############################################################################
    np_triangle_raster = np_triangle_rasterised(size)
    ############################################################################
    # cull face is set to back, only one 'front' triangle in the scene => no culling
    _do_test_cull_face('back', np_triangle_raster)

    ############################################################################
    # EXPECTED DATAS
    # It should have 1.0's everywhere.
    ############################################################################
    np_clear_depth = np.full(size, 1.0)
    ############################################################################
    # cull face is set to front, only one 'front' triangle in the scene => the triangle is culled
    _do_test_cull_face('front', np_clear_depth)

    # cull face is set to front_and_back, only one 'front' triangle in the scene => the triangle is culled
    _do_test_cull_face('front_and_back', np_clear_depth)

    ctx.enable_only(mgl.NOTHING)



import struct


def test_1(ctx):
    rbo1 = ctx.renderbuffer((4, 4), dtype='f1')
    rbo2 = ctx.renderbuffer((4, 4))
    rbo3 = ctx.renderbuffer((4, 4))
    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer(rbo2)
    fbo3 = ctx.framebuffer(rbo3)
    fbo4 = ctx.framebuffer([rbo1, rbo2, rbo3])

    fbo1.clear(1.0, 0.0, 0.0, 1.0)
    fbo2.clear(0.0, 1.0, 0.0, 1.0)
    fbo3.clear(0.0, 0.0, 1.0, 1.0)

    buf = ctx.buffer(reserve=64)

    fbo4.read_into(buf, attachment=0, components=4, dtype='f1')
    assert buf.read() == b'\xff\x00\x00\xff' * 16

    fbo4.read_into(buf, attachment=1, components=4, dtype='f1')
    assert buf.read() == b'\x00\xff\x00\xff' * 16

    fbo4.read_into(buf, attachment=2, components=4, dtype='f1')
    assert buf.read() == b'\x00\x00\xff\xff' * 16


def test_2(ctx):
    pixels = b'\xFF\x00\x00\xFF\x00\xFF\x00\xFF\x00\x00\xFF\xFF\xFF\xFF\xFF\xFF'

    texture = ctx.texture((2, 2), 4)
    buf = ctx.buffer(pixels)
    texture.write(buf)

    assert texture.read() == pixels


def test_3(ctx):
    pixels = struct.pack(
        '4f4f4f4f',
        1.0, 0.0, 0.0, 1.0,
        0.0, 1.0, 0.0, 1.0,
        0.0, 0.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0,
    )

    buf = ctx.buffer(pixels)
    texture = ctx.texture((2, 2), 4, dtype='f4')
    texture.write(buf)

    assert texture.read() == pixels


def test_4(ctx):
    pixels = b'\xFF\x00\x00\xFF\x00\xFF\x00\xFF\x00\x00\xFF\xFF\xFF\xFF\xFF\xFF'

    texture = ctx.texture((2, 2), 4, pixels)
    buf = ctx.buffer(reserve=len(pixels))
    texture.read_into(buf)

    assert buf.read() == pixels


def test_5(ctx):
    pixels = struct.pack(
        '4f4f4f4f',
        1.0, 0.0, 0.0, 1.0,
        0.0, 1.0, 0.0, 1.0,
        0.0, 0.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0,
    )

    texture = ctx.texture((2, 2), 4, pixels, dtype='f4')
    buf = ctx.buffer(reserve=len(pixels))
    texture.read_into(buf)

    assert buf.read() == pixels



import struct
import pytest
import moderngl


@pytest.fixture(scope='module')
def prog(ctx_static):
    return ctx_static.program(
        vertex_shader='''
            #version 330

            in float v_in_1;
            in float v_in_2;
            in float v_in_3;
            out float v_out;

            void main() {
                v_out = v_in_1 + v_in_2 + v_in_3;
            }
        ''',
        varyings=['v_out']
    )


@pytest.fixture(scope='module')
def vbo1(ctx_static):
    return ctx_static.buffer(struct.pack('4f', 1.0, 2.0, 3.0, 4.0))


@pytest.fixture(scope='module')
def vbo2(ctx_static):
    return ctx_static.buffer(struct.pack('4f', 10.0, 20.0, 30.0, 40.0))


@pytest.fixture(scope='module')
def vbo3(ctx_static):
    return ctx_static.buffer(struct.pack('4f', 100.0, 200.0, 300.0, 400.0))


@pytest.fixture(scope='module')
def vao1(ctx_static, prog, vbo1, vbo2):
    content = [
        (vbo1, 'f', 'v_in_1'),
        (vbo2, 'f/i', 'v_in_2'),
    ]
    return ctx_static.vertex_array(prog, content)


@pytest.fixture(scope='module')
def vao2(ctx_static, prog, vbo1, vbo2, vbo3):
    content = [
        (vbo1, 'f', 'v_in_1'),
        (vbo2, 'f/i', 'v_in_2'),
        (vbo3, 'f/r', 'v_in_3'),
    ]
    return ctx_static.vertex_array(prog, content)


@pytest.fixture(scope='module')
def res(ctx_static):
    return ctx_static.buffer(reserve=1024)


def test_vertex_attrib_per_instance_1(vao1, res):
    vao1.transform(res, moderngl.POINTS, vertices=4, instances=1)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 11.0
    assert pytest.approx(b) == 12.0
    assert pytest.approx(c) == 13.0
    assert pytest.approx(d) == 14.0


def test_vertex_attrib_per_instance_2(vao1, res):
    vao1.transform(res, moderngl.POINTS, vertices=2, instances=2)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 11.0
    assert pytest.approx(b) == 12.0
    assert pytest.approx(c) == 21.0
    assert pytest.approx(d) == 22.0


def test_vertex_attrib_per_instance_3(vao1, res):
    vao1.transform(res, moderngl.POINTS, vertices=1, instances=4)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 11.0
    assert pytest.approx(b) == 21.0
    assert pytest.approx(c) == 31.0
    assert pytest.approx(d) == 41.0


def test_vertex_attrib_per_render_1(vao2, res):
    vao2.transform(res, moderngl.POINTS, vertices=4, instances=1)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 111.0
    assert pytest.approx(b) == 112.0
    assert pytest.approx(c) == 113.0
    assert pytest.approx(d) == 114.0


def test_vertex_attrib_per_render_2(vao2, res):
    vao2.transform(res, moderngl.POINTS, vertices=2, instances=2)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 111.0
    assert pytest.approx(b) == 112.0
    assert pytest.approx(c) == 121.0
    assert pytest.approx(d) == 122.0


def test_vertex_attrib_per_render_3(vao2, res):
    vao2.transform(res, moderngl.POINTS, vertices=1, instances=4)

    a, b, c, d = struct.unpack('4f', res.read(16))

    assert pytest.approx(a) == 111.0
    assert pytest.approx(b) == 121.0
    assert pytest.approx(c) == 131.0
    assert pytest.approx(d) == 141.0



import numpy as np
import numpy.testing as npt
import moderngl


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec2 in_vert;
            out vec2 v_text;

            void main() {
                v_text = in_vert / 2.0 + 0.5;
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330

            uniform sampler2D Texture;

            in vec2 v_text;
            out vec4 f_color;

            void main() {
                f_color = texture(Texture, v_text);
            }
        ''',
    )

    pixels = np.tile([0.0, 0.5, 1.0], 4 * 4).astype('f2')
    tex = ctx.texture((4, 4), 3, pixels, dtype='f2')

    vbo = ctx.buffer(np.array([-1, -1, -1, 1, 1, -1, 1, 1], dtype='f4'))
    vao = ctx.simple_vertex_array(prog, vbo, 'in_vert')
    fbo = ctx.simple_framebuffer((4, 4))

    fbo.use()
    fbo.clear()
    tex.use()
    vao.render(moderngl.TRIANGLE_STRIP)
    res = np.frombuffer(fbo.read(dtype='f2'), dtype='f2')
    npt.assert_almost_equal(res, pixels, decimal=2)



import moderngl
import pytest


def test_properties(ctx):
    tex = ctx.texture3d((2, 4, 8), 4)
    assert tex.size == (2, 4, 8)
    assert tex.width == 2
    assert tex.height == 4
    assert tex.depth == 8
    assert tex.components == 4
    assert tex.dtype == 'f1'
    assert tex.glo > 0
    assert tex.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert tex.repeat_x is True
    assert tex.repeat_y is True
    assert tex.repeat_z is True
    assert tex == tex

    tex.repeat_x = False
    tex.repeat_y = False
    tex.repeat_z = False
    assert tex.repeat_x is False
    assert tex.repeat_y is False
    assert tex.repeat_z is False


def test_mipmaps(ctx):
    tex = ctx.texture3d((8, 8, 8), 4)
    tex.build_mipmaps(max_level=2)
    tex.filter = moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR


def test_texture_3d_create_1(ctx):
    ctx.texture3d((8, 8, 8), 3)


def test_texture_3d_create_2(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8 * 8
    ctx.texture3d((8, 8, 8), 3, pixels)


def test_texture_3d_create_string(ctx):
    pixels = 'abc' * 8 * 8
    with pytest.raises(Exception):
        ctx.texture3d((8, 8, 8), 3, pixels)


def test_texture_3d_create_wrong_size(ctx):
    with pytest.raises(Exception):
        ctx.texture3d((8, 8), 3)


def test_texture_3d_get_swizzle(ctx):
    tex = ctx.texture3d((8, 8, 8), 4)
    assert tex.swizzle == 'RGBA'


def test_texture_3d_swizzle_1(ctx):
    tex = ctx.texture3d((8, 8, 8), 4)
    tex.swizzle = 'argb'
    assert tex.swizzle == 'ARGB'


def test_texture_3d_swizzle_2(ctx):
    tex = ctx.texture3d((8, 8, 8), 1)
    tex.swizzle = 'RRRR'
    assert tex.swizzle == 'RRRR'


def test_texture_3d_swizzle_3(ctx):
    tex = ctx.texture3d((8, 8, 8), 2)
    tex.swizzle = '01RG'
    assert tex.swizzle == '01RG'


def test_texture_3d_read(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8 * 8
    tex = ctx.texture3d((8, 8, 8), 3, pixels)
    assert tex.read() == pixels


def test_texture_3d_read_into(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8 * 8
    tex = ctx.texture3d((8, 8, 8), 3, pixels)
    buf = bytearray(8 * 8 * 8 * 3)
    tex.read_into(buf)
    assert bytes(buf) == pixels


def test_texture_3d_read_into_pbo(ctx):
    if ctx.info['GL_VENDOR'].startswith('Intel'):
        raise pytest.skip('Intel drivers do not support PBO')

    pixels = b'\x10\x20\x30' * 8 * 8 * 8
    tex = ctx.texture3d((8, 8, 8), 3, pixels)
    buf = ctx.buffer(reserve=8 * 8 * 8 * 3)
    tex.read_into(buf)
    assert buf.read() == pixels


def test_texture_3d_write_1(ctx):
    pixels1 = b'\x00\x00\x00' * 8 * 8 * 8
    pixels2 = b'\xff\xff\xff' * 8 * 8 * 8

    tex = ctx.texture3d((8, 8, 8), 3, pixels1)
    assert tex.read() == pixels1

    tex.write(pixels2)
    assert tex.read() == pixels2


def test_texture_3d_write_2(ctx):
    pixels1 = b'\x00\x00\x00' * 8 * 8 * 8
    pixels2 = b'\xff\xff\xff' * 6 * 6 * 6
    pixels3 = b'\x10\x20\x30'

    tex = ctx.texture3d((8, 8, 8), 3)

    tex.write(pixels1, viewport=(0, 0, 0, 8, 8, 8))
    tex.write(pixels2, viewport=(1, 1, 1, 6, 6, 6))
    tex.write(pixels3, viewport=(2, 4, 6, 1, 1, 1))

    def pixel(x, y, z):
        if x == 2 and y == 4 and z == 6:
            return b'\x10\x20\x30'

        if x == 0 or y == 0 or z == 0 or x == 7 or y == 7 or z == 7:
            return b'\x00\x00\x00'

        return b'\xff\xff\xff'

    expectation = b''.join(pixel(x, y, z) for z in range(8) for y in range(8) for x in range(8))
    assert tex.read() == expectation


def test_texture_default_filter(ctx):
    """Ensure default filter is correct"""
    # Float types
    for dtype in ["f1", "f2", "f4"]:
        texture = ctx.texture3d((10, 10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)

    for dtype in ["u1", "u2", "u4", "i1", "i2", "i4"]:
        texture = ctx.texture3d((10, 10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.NEAREST, moderngl.NEAREST)


def test_bind_to_image(ctx):
    if ctx.version_code < 430:
        pytest.skip('compute shaders not supported')

    tex = ctx.texture3d((4, 4, 4), 4)
    tex.bind_to_image(0, read=True, write=False)
    tex.bind_to_image(1, read=False, write=True)




def test_viewport(ctx):
    fbo = ctx.simple_framebuffer((10, 10))
    fbo.use()
    assert ctx.viewport == (0, 0, 10, 10)
    assert fbo.viewport == (0, 0, 10, 10)
    ctx.viewport = 2, 3, 7, 8
    assert fbo.viewport == (2, 3, 7, 8)
    assert ctx.viewport == (2, 3, 7, 8)
    assert ctx.fbo.viewport == (2, 3, 7, 8)




VERTEX_SHADER = '''
    #version 330
    uniform mat4 Mvp;
    in vec4 vertex;
    uniform Block {
        mat4 Transform;
    };
    void main() {
        gl_Position = Mvp * Transform * vertex;
    }
'''

GEOMETRY_SHADER = '''
    #version 330
    layout(points) in;
    uniform vec4 Position;
    layout(line_strip, max_vertices = 2) out;
    out vec4 color;
    void main() {
        gl_Position = gl_in[0].gl_Position + Position;
        EmitVertex();
        gl_Position = gl_in[0].gl_Position - Position;
        EmitVertex();
        EndPrimitive();
    }
'''

FRAGMENT_SHADER = '''
    #version 330
    uniform vec4 Color;
    uniform Lights {
        uniform vec4 Light[1];
    };
    out vec4 color;
    void main() {
        color = Color * Light[0];
    }
'''


def test_program_uniforms_1(ctx):
    program = ctx.program(vertex_shader=VERTEX_SHADER)

    assert 'Mvp' in program
    assert 'Block' in program
    assert 'Transform' not in program


def test_program_uniforms_2(ctx):
    program = ctx.program(
        vertex_shader=VERTEX_SHADER,
        geometry_shader=GEOMETRY_SHADER,
        fragment_shader=FRAGMENT_SHADER,
    )

    assert 'Mvp' in program
    assert 'Position' in program
    assert 'Color' in program


def test_program_uniform_blocks(ctx):
    program = ctx.program(
        vertex_shader=VERTEX_SHADER,
        fragment_shader=FRAGMENT_SHADER,
    )

    assert 'Block' in program
    assert 'Lights' in program
    assert 'Transform' not in program
    assert 'Light' not in program


def test_program_geometry_primitives_1(ctx):
    program = ctx.program(vertex_shader=VERTEX_SHADER)

    assert program.geometry_input is None
    assert program.geometry_output is None
    assert program.geometry_vertices == 0


def test_program_geometry_primitives_2(ctx):
    program = ctx.program(
        vertex_shader=VERTEX_SHADER,
        geometry_shader=GEOMETRY_SHADER,
    )
    assert program.geometry_vertices == 2



import struct
import pytest


@pytest.fixture(scope='module', autouse=True)
def prog(ctx_static):
    return ctx_static.program(
        vertex_shader='''
            #version 330

            in vec2 in_v;
            out vec2 out_v;

            uniform Block1 {
                float x;
            };

            uniform Block2 {
                float y;
            };

            uniform Block3 {
                float z;
            };

            void main() {
                out_v = in_v * z + vec2(x, y);
            }
        ''',
        varyings=['out_v']
    )


def test_1(ctx, prog):
    buf_v = ctx.buffer(struct.pack('2f', 100.0, 1000.0))
    buf_u1 = ctx.buffer(struct.pack('f', 9.5))
    buf_u2 = ctx.buffer(struct.pack('f', 4.0))
    buf_u3 = ctx.buffer(struct.pack('f', 3.0))
    buf_u4 = ctx.buffer(struct.pack('f', 0.0))
    buf_r = ctx.buffer(reserve=buf_v.size)

    vao = ctx.vertex_array(prog, [
        (buf_v, '2f', 'in_v'),
    ])

    assert 'Block1' in prog
    assert 'Block2' in prog
    assert 'Block3' in prog

    assert 'x' not in prog
    assert 'y' not in prog
    assert 'z' not in prog

    prog['Block1'].binding = 2
    prog['Block2'].binding = 4
    prog['Block3'].binding = 1

    buf_u1.bind_to_uniform_block(2)
    buf_u2.bind_to_uniform_block(4)
    buf_u4.bind_to_uniform_block(1)

    vao.transform(buf_r)
    a, b = struct.unpack('2f', buf_r.read())
    assert pytest.approx(a) == 9.5
    assert pytest.approx(b) == 4.0

    buf_u3.bind_to_uniform_block(1)

    vao.transform(buf_r)
    a, b = struct.unpack('2f', buf_r.read())
    assert pytest.approx(a) == 309.5
    assert pytest.approx(b) == 3004.0


def test_2(ctx, prog):
    min_offset = ctx.info['GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT']
    u3_len = (min_offset // 4) + 1
    u3_data = list(range(u3_len))

    buf_v = ctx.buffer(struct.pack('2f', 100.0, 1000.0))
    buf_u1 = ctx.buffer(struct.pack('f', 9.5))
    buf_u2 = ctx.buffer(struct.pack('f', 4.0))
    buf_u3 = ctx.buffer(struct.pack('{:d}f'.format(u3_len), *u3_data))
    buf_r = ctx.buffer(reserve=buf_v.size)

    vao = ctx.vertex_array(prog, [
        (buf_v, '2f', 'in_v'),
    ])

    prog['Block1'].binding = 2
    prog['Block2'].binding = 4
    prog['Block3'].binding = 1

    buf_u1.bind_to_uniform_block(2)
    buf_u2.bind_to_uniform_block(4)
    buf_u3.bind_to_uniform_block(1, offset=min_offset)

    vao.transform(buf_r)
    a, b = struct.unpack('2f', buf_r.read())
    assert pytest.approx(a) == u3_data[-1] * 100 + 9.5
    assert pytest.approx(b) == u3_data[-1] * 1000 + 4.0

    buf_u3.bind_to_uniform_block(1, offset=0)

    vao.transform(buf_r)
    a, b = struct.unpack('2f', buf_r.read())
    assert pytest.approx(a) == 9.5
    assert pytest.approx(b) == 4.0



import pytest


def test_1(ctx):
    buf = ctx.buffer(b'abc')

    with pytest.raises(Exception):
        buf.read(4)

    with pytest.raises(Exception):
        buf.read(offset=-1)

    with pytest.raises(Exception):
        buf.read(offset=1, size=3)


def test_2(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.read_chunks(1, 4, 1, 6)

    with pytest.raises(Exception):
        buf.read_chunks(1, 4, -1, 6)

    with pytest.raises(Exception):
        buf.read_chunks(2, -1, 2, 1)


def test_3(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.read_chunks(2, 2, 3, 3)

    with pytest.raises(Exception):
        buf.read_chunks(2, -1, -3, 3)

    with pytest.raises(Exception):
        buf.read_chunks(2, -4, -3, 3)


def test_4(ctx):
    buf = ctx.buffer(b'123456789')

    with pytest.raises(Exception):
        buf.read_chunks(3, 0, 2, 2)



import moderngl
import pytest


def test_buffer_create_bytes(ctx):
    ctx.buffer(b'Hello World!')


def test_buffer_reserve(ctx):
    ctx.buffer(reserve=1024)


def test_buffer_reserve_human_readable(ctx):
    buf = ctx.buffer(reserve='2KB')
    assert buf.size == 2 * 1024


def test_buffer_data_and_reserve(ctx):
    with pytest.raises(moderngl.Error):
        ctx.buffer(b'Hello World!', reserve=1024)


def test_data_size(ctx):
    buf = ctx.buffer(b'\x00\x00\x00\x00')
    assert buf.size == 4


def test_reserve_size(ctx):
    buf = ctx.buffer(reserve=8)
    assert buf.size == 8


def test_dynamic(ctx):
    buf = ctx.buffer(b'\x00\x00\x00\x00', dynamic=False)
    assert buf.dynamic is False


def test_non_dynamic(ctx):
    buf = ctx.buffer(b'\x00\x00\x00\x00', dynamic=True)
    assert buf.dynamic is True



import moderngl
import pytest


def test_texture_create_1(ctx):
    ctx.texture((16, 16), 3)


def test_texture_create_2(ctx):
    pixels = b'\x10\x20\x30' * 16 * 16
    ctx.texture((16, 16), 3, pixels)


def test_texture_create_string(ctx):
    pixels = 'abc' * 16 * 16
    with pytest.raises(Exception):
        ctx.texture((16, 16), 3, pixels)


def test_multisample_texture(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    ctx.texture((16, 16), 3, samples=2)


def test_depth_texture(ctx):
    ctx.depth_texture((16, 16))


def test_multisample_depth_texture(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    ctx.depth_texture((16, 16), samples=2)


def test_texture_invalid_samples(ctx):
    if ctx.max_samples < 2:
        pytest.skip('multisampling is not supported')

    with pytest.raises(moderngl.Error, match='sample'):
        ctx.texture((16, 16), 3, samples=3)


def test_texture_get_swizzle(ctx):
    tex = ctx.texture((16, 16), 4)
    assert tex.swizzle == 'RGBA'


def test_texture_swizzle_1(ctx):
    tex = ctx.texture((16, 16), 4)
    tex.swizzle = 'argb'
    assert tex.swizzle == 'ARGB'


def test_texture_swizzle_2(ctx):
    tex = ctx.texture((16, 16), 1)
    tex.swizzle = 'RRRR'
    assert tex.swizzle == 'RRRR'


def test_texture_swizzle_3(ctx):
    tex = ctx.texture((16, 16), 2)
    tex.swizzle = '01RG'
    assert tex.swizzle == '01RG'


def test_texture_read(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8
    tex = ctx.texture((8, 8), 3, pixels)
    assert tex.read() == pixels


def test_texture_read_into(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8
    tex = ctx.texture((8, 8), 3, pixels)
    buf = bytearray(8 * 8 * 3)
    tex.read_into(buf)
    assert bytes(buf) == pixels


def test_texture_read_into_pbo(ctx):
    pixels = b'\x10\x20\x30' * 8 * 8
    tex = ctx.texture((8, 8), 3, pixels)
    buf = ctx.buffer(reserve=8 * 8 * 3)

    tex.read_into(buf)
    assert buf.read() == pixels


def test_texture_write_1(ctx):
    pixels1 = b'\x00\x00\x00' * 8 * 8
    pixels2 = b'\xff\xff\xff' * 8 * 8

    tex = ctx.texture((8, 8), 3, pixels1)
    assert tex.read() == pixels1

    tex.write(pixels2)
    assert tex.read() == pixels2


def test_texture_write_2(ctx):
    pixels1 = b'\x00\x00\x00' * 8 * 8
    pixels2 = b'\xff\xff\xff' * 6 * 6
    pixels3 = b'\x10\x20\x30'

    tex = ctx.texture((8, 8), 3)

    tex.write(pixels1, viewport=(0, 0, 8, 8))
    tex.write(pixels2, viewport=(1, 1, 6, 6))
    tex.write(pixels3, viewport=(2, 4, 1, 1))

    def pixel(x, y):
        if x == 2 and y == 4:
            return b'\x10\x20\x30'

        if x == 0 or y == 0 or x == 7 or y == 7:
            return b'\x00\x00\x00'

        return b'\xff\xff\xff'

    expectation = b''.join(pixel(x, y) for y in range(8) for x in range(8))
    assert tex.read() == expectation


def test_texture_alignment_1(ctx):
    tex = ctx.texture((3, 3), 3)
    assert len(tex.read(alignment=1)) == 27


def test_texture_alignment_2(ctx):
    tex = ctx.texture((3, 3), 3)
    assert len(tex.read(alignment=2)) == 30


def test_texture_alignment_3(ctx):
    tex = ctx.texture((3, 3), 3)
    assert len(tex.read(alignment=4)) == 36


def test_texture_alignment_4(ctx):
    tex = ctx.texture((3, 3), 3)
    assert len(tex.read(alignment=8)) == 48


def test_texture_alignment_5(ctx):
    pixels = b'\x80\x90\xA0\x80\x90\xA0\x80\x90\xA0\x80\x90\xA0'
    tex = ctx.texture((2, 2), 3, pixels, alignment=1)
    assert tex.read(alignment=1) == pixels


def test_texture_alignment_6(ctx):
    pixels = b'\x80\x90\xA0\x80\x90\xA0\x00\x00\x80\x90\xA0\x80\x90\xA0\x00\x00'
    tex = ctx.texture((2, 2), 3, pixels, alignment=4)
    result = tex.read(alignment=4)
    assert result[0:6] == b'\x80\x90\xA0\x80\x90\xA0'
    assert result[8:14] == b'\x80\x90\xA0\x80\x90\xA0'



import moderngl


def test_shader_includes(ctx):
    ctx.includes['uniforms'] = '''
        uniform vec2 pos;
        uniform float scale;
    '''

    program = ctx.program(
        vertex_shader='''
            #version 330

            #include "uniforms"

            in vec2 vert;
            out vec2 v_vert;

            void main() {
                gl_Position = vec4(pos + vert * scale, 0.0, 1.0);
                v_vert = vert;
            }
        ''',
        fragment_shader='''
            #version 330

            in vec2 v_vert;
            out vec4 color;

            void main() {
                color = vec4(v_vert, 0.0, 1.0);
            }
        ''',
    )
    assert program.glo > 0

    assert 'vert' in program
    assert 'pos' in program
    assert 'scale' in program
    assert program.is_transform is False

    assert isinstance(program['vert'], moderngl.Attribute)
    assert isinstance(program['pos'], moderngl.Uniform)
    assert isinstance(program['scale'], moderngl.Uniform)



import pytest
import numpy as np
import moderngl
import moderngl as mgl
from scipy import ndimage

# https://docs.oracle.com/cd/E19957-01/806-3568/ncg_goldberg.html
# https://en.wikipedia.org/wiki/Machine_epsilon
# binary32	single precision	float	2	24 (one bit is implicit)	2−24 ≈ 5.96e-08	2−23 ≈ 1.19e-07
EPSILON_IEEE_754 = 5.96e-08


@pytest.fixture
def prog_render_depth_pass(ctx):
    return ctx.program(
        vertex_shader='''
                #version 330

                in vec3 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 1.0);
                }
            ''',
        fragment_shader='''
                #version 330

                void main() {
                }
            ''',
    )


@pytest.fixture
def prog_draw_depth(ctx):
    prog = ctx.program(
        vertex_shader='''
                    #version 330

                    in vec2 in_vert;
                    out vec2 uv;

                    void main() {
                        uv = in_vert;
                        gl_Position = vec4(in_vert * 2.0 - 1.0, 0.0, 1.0);
                    }
                ''',
        # from `texture.py`
        fragment_shader='''
                    #version 330

                    uniform sampler2D depth;

                    in vec2 uv;
                    out float fragColor;

                    void main() {
                        float raw_depth_nonlinear = texture(depth, uv).r;
                        fragColor = raw_depth_nonlinear;
                    }
                ''',
    )
    prog['depth'].value = 0
    return prog


@pytest.fixture
def vbo_triangle(ctx):
    vertices = np.array([[-1, 1, 0.], [-1, -1, 0.], [1, -1, 0.], ])
    return ctx.buffer(vertices.astype('f4').tobytes())


@pytest.fixture
def vbo_quad(ctx):
    canvas_fs_quad = np.array(
        [0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0]).astype('f4')
    return ctx.buffer(canvas_fs_quad.tobytes())


@pytest.fixture
def prog_shadow_test(ctx):
    prog = ctx.program(
        vertex_shader='''
                #version 330

                in vec2 in_vert;
                out vec2 uv;

                void main() {
                    uv = in_vert;
                    gl_Position = vec4(in_vert * 2.0 - 1.0, 0.0, 1.0);
                }
        ''',
        fragment_shader='''
                #version 330

                uniform sampler2DShadow depth;

                uniform vec2 u_shadow_coord_offset;

                uniform float u_depth_bias;
                uniform float u_depth_triangle;

                in vec2 uv;
                out float fragColor;

                void main() {
                    // Compute depth caster
                    float depth_shadow_caster = u_depth_triangle + u_depth_bias;
                    // Compute shadow coordinates + depth of caster
                    vec3 shadow_coord = vec3(uv, depth_shadow_caster);
                    shadow_coord += vec3(u_shadow_coord_offset, 0.0);
                    // Compute visibility:
                    // -> shadow test: receiver (potential) vs caster
                    float visibility = texture(depth, shadow_coord);
                    // Output the visibility
                    fragColor = visibility;
                }
            ''',
    )
    #
    prog['u_depth_bias'].value = EPSILON_IEEE_754
    depth_triangle_in_obj_space = 0
    depth_triangle_in_view_space = (depth_triangle_in_obj_space + 1) * 0.5
    prog['u_depth_triangle'].value = depth_triangle_in_view_space
    #
    return prog


@pytest.fixture
def fbo_with_rasterized_triangle(ctx, vbo_triangle):
    def _build_fbo_with_rasterized_triangle(prog, size=(4, 4), depth_clear=1.0):
        tex_depth = ctx.depth_texture(size)  # implicit -> dtype='f4', components=1
        fbo_depth = ctx.framebuffer(depth_attachment=tex_depth)
        fbo_depth.use()
        fbo_depth.clear(depth=depth_clear)

        # vertex array object of triangle with depth pass prog
        vao_triangle = ctx.simple_vertex_array(prog, vbo_triangle, 'in_vert')
        # Now we render a triangle in there
        vao_triangle.render()

        return fbo_depth, tex_depth

    return _build_fbo_with_rasterized_triangle


@pytest.fixture
def np_triangle_rasterized():
    def _build_np_triangle_rasterized(size):
        # It should have 0.5's where the triangle lies.
        depth_value_from_triangle_vertices = 0.0
        # Map [-1, +1] -> [0, 1]
        depth_value_in_depth_buffer = (depth_value_from_triangle_vertices + 1) * 0.5
        return np.array(
            [
                [depth_value_in_depth_buffer] * (size[0] - (j + 1)) + [1.0] * (j + 1)
                for j in range(size[1])
            ]
        )

    return _build_np_triangle_rasterized


def test_depth_sampler(
    ctx,
    prog_render_depth_pass,
    prog_draw_depth,
    vbo_triangle,
    vbo_quad,
    fbo_with_rasterized_triangle,
    np_triangle_rasterized,
):
    ctx.enable(mgl.DEPTH_TEST)

    size = (16,) * 2
    fbo_depth, tex_depth = fbo_with_rasterized_triangle(prog_render_depth_pass, size)
    depth_from_dbo = np.frombuffer(tex_depth.read(), dtype=np.dtype('f4')).reshape(size[::-1])

    ############################################################################
    # EXPECTED DATAS
    # It should have 0.5's where the triangle lies.
    ############################################################################
    np_triangle_raster = np_triangle_rasterized(size)

    depth_tex_filter = (moderngl.NEAREST, moderngl.NEAREST)
    compare_func = ''

    def _do_test():
        # some stuff should be 1s and 0.5 where the triangle lies.
        np.testing.assert_array_almost_equal(depth_from_dbo, np_triangle_raster)

        fbo_draw_depth = ctx.framebuffer(color_attachments=[ctx.renderbuffer(size, components=1, dtype='f4')])
        fbo_draw_depth.use()

        tex_depth.use(location=0)

        # vertex array object of (fullscreen) quad with render depth prog
        vao_quad = ctx.simple_vertex_array(prog_draw_depth, vbo_quad, 'in_vert')
        vao_quad.render(moderngl.TRIANGLE_STRIP)

        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(data, np_triangle_raster)

    def _with_sampler():
        # from: https://github.com/Contraz/demosys-py/blob/master/demosys/opengl/texture.py
        depth_sampler = ctx.sampler(filter=depth_tex_filter, compare_func=compare_func)  # disable depth func
        depth_sampler.use(location=0)
        _do_test()

    def _with_texture_parameters():
        tex_depth.filter = depth_tex_filter
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_parameters()


def test_sampler_shadow(
    ctx,
    prog_render_depth_pass, prog_shadow_test,
    vbo_triangle, vbo_quad,
    fbo_with_rasterized_triangle,
):
    ctx.enable(mgl.DEPTH_TEST)

    size = (3,) * 2
    fbo_depth, tex_depth = fbo_with_rasterized_triangle(prog_render_depth_pass, size)

    ############################################################################
    # EXPECTED DATAS                                                                                                   #
    # It should have 1's where the triangle lies.                                                                      #
    ############################################################################
    texel_shadowed = 1
    np_triangle_raster_visibility = np.array(
        [
            [texel_shadowed] * (size[0] - (j + 1)) + [1 - texel_shadowed] * (j + 1)
            for j in range(size[1])
        ]
    )
    ############################################################################

    # No offset - fetch the center pixel (NEAREST)
    prog_shadow_test['u_shadow_coord_offset'].value = (0,) * 2

    depth_tex_filter = (moderngl.NEAREST, moderngl.NEAREST)
    compare_func = '>='

    def _do_test():
        fbo_draw_depth = ctx.framebuffer(color_attachments=[ctx.renderbuffer(size, components=1, dtype='f4')])
        fbo_draw_depth.use()

        tex_depth.use(location=0)

        # https://www.khronos.org/opengl/wiki/Sampler_(GLSL)#Shadow_samplers
        vao_fs_shadow = ctx.simple_vertex_array(prog_shadow_test, vbo_quad, 'in_vert')
        vao_fs_shadow.render(moderngl.TRIANGLE_STRIP)

        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'), dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(data, np_triangle_raster_visibility)

    def _with_sampler():
        # https://github.com/Contraz/demosys-py/blob/01f285cd3a132012e14a51bf0ae9d7aa5a489b55/demosys/opengl/texture.py#L151
        shadow_sampler = ctx.sampler(filter=depth_tex_filter, compare_func=compare_func)  # enable depth func
        shadow_sampler.use(location=0)
        _do_test()
        shadow_sampler.clear(location=0)

    def _with_texture_parameters():
        tex_depth.filter = depth_tex_filter
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_parameters()


def test_sampler_shadow_with_bilinear_interpolation(
        ctx,
        prog_render_depth_pass, prog_shadow_test,
        vbo_triangle, vbo_quad,
        fbo_with_rasterized_triangle
):
    ctx.enable(mgl.DEPTH_TEST)
    size = (7,) * 2

    fbo_depth, tex_depth = fbo_with_rasterized_triangle(prog_render_depth_pass, size)

    ############################################################################
    # EXPECTED DATAS                                                           #
    ############################################################################
    # Box filter (2x2) on triangle visibility rasterization
    # is similar to bilinear interpolation with samplerShadow
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.uniform_filter.html#scipy.ndimage.uniform_filter
    np_triangle_raster_visibility = np.array(
        [
            [1.0] * (size[0] - (j + 1)) + [0.0] * (j + 1)
            for j in range(size[1])
        ]
    )
    box_filter_on_triangle_raster_visibility = ndimage.uniform_filter(
        np_triangle_raster_visibility,
        size=2, mode='constant', cval=0, origin=-1)
    ############################################################################

    half_pixel_size = (1.0 / size[0] * 0.5, 1.0 / size[1] * 0.5)
    prog_shadow_test['u_shadow_coord_offset'].value = tuple(half_pixel_size)

    def _do_test():
        fbo_draw_depth = ctx.framebuffer(
            color_attachments=[ctx.renderbuffer(size,
                                                components=1, dtype='f4')])
        fbo_draw_depth.use()
        tex_depth.use(location=0)

        # https://www.khronos.org/opengl/wiki/Sampler_(GLSL)#Shadow_samplers
        vao_fs_shadow = ctx.simple_vertex_array(prog_shadow_test, vbo_quad,
                                                'in_vert')
        vao_fs_shadow.render(moderngl.TRIANGLE_STRIP)

        #
        data = np.frombuffer(fbo_draw_depth.read(components=1, dtype='f4'),
                             dtype=np.dtype('f4')).reshape(size[::-1])
        np.testing.assert_array_almost_equal(
            data, box_filter_on_triangle_raster_visibility)

    depth_tex_filter = (moderngl.LINEAR, moderngl.LINEAR)
    compare_func = '>='

    def _with_sampler():
        # https://github.com/Contraz/demosys-py/blob/01f285cd3a132012e14a51bf0ae9d7aa5a489b55/demosys/opengl/texture.py#L151
        shadow_sampler = ctx.sampler(
            filter=depth_tex_filter,
            compare_func=compare_func,  # enable depth func
            # from context.py:
            #   border_color (tuple): The (r, g, b, a) color for the texture border.
            #   When this value is set the ``repeat_`` values are overridden
            #   setting the texture wrap to return the border color when outside ``[0, 1]`` range.
            # border_color=(0, 0, 0, 0),    # seems not to work well ...
            repeat_x=False, repeat_y=False,
        )
        shadow_sampler.use(location=0)
        _do_test()
        shadow_sampler.clear(location=0)

    def _with_texture_parameters():
        tex_depth.filter = depth_tex_filter
        tex_depth.repeat_x = False
        tex_depth.repeat_y = False
        tex_depth.compare_func = compare_func
        _do_test()

    _with_sampler()
    _with_texture_parameters()



import moderngl


def test_create(ctx):
    sampler = ctx.sampler()
    sampler.use(location=0)
    sampler.clear(location=0)


def test_defaults(ctx):
    sampler = ctx.sampler()
    assert sampler.anisotropy <= 1.0
    assert sampler.repeat_x is True
    assert sampler.repeat_y is True
    assert sampler.repeat_z is True
    assert sampler.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert sampler.compare_func == '?'
    assert sampler.border_color == (0.0, 0.0, 0.0, 0.0)
    assert sampler.min_lod == -1000.0
    assert sampler.max_lod == 1000.0


def test_prop_changes(ctx):
    sampler = ctx.sampler()
    # Change values
    sampler.anisotropy = ctx.max_anisotropy
    sampler.filter = (moderngl.NEAREST_MIPMAP_NEAREST, moderngl.NEAREST)
    sampler.compare_func = "<="

    assert sampler.anisotropy == ctx.max_anisotropy
    assert sampler.filter == (moderngl.NEAREST_MIPMAP_NEAREST, moderngl.NEAREST)
    assert sampler.compare_func == "<="

    # Ensure repeat parameters are set correctly
    sampler.repeat_x = False
    assert (sampler.repeat_x, sampler.repeat_y, sampler.repeat_z) == (False, True, True)
    sampler.repeat_y = False
    assert (sampler.repeat_x, sampler.repeat_y, sampler.repeat_z) == (False, False, True)
    sampler.repeat_z = False
    assert (sampler.repeat_x, sampler.repeat_y, sampler.repeat_z) == (False, False, False)


def test_border_color(ctx):
    sampler = ctx.sampler()

    # Ensure border color values are set correctly
    colors = [
        (1.0, 0.0, 0.0, 0.0),
        (0.0, 1.0, 0.0, 0.0),
        (0.0, 0.0, 1.0, 0.0),
        (0.0, 0.0, 0.0, 1.0),
    ]
    for color in colors:
        sampler.border_color = color
        assert sampler.border_color == color


def test_lod(ctx):
    sampler = ctx.sampler()

    sampler.min_lod = 0.0
    assert sampler.min_lod == 0.0

    sampler.max_lod = 500.0
    assert sampler.max_lod == 500.0


def test_clear_samplers(ctx):
    ctx.clear_samplers(start=0, end=5)
    ctx.clear_samplers(start=5, end=10)
    ctx.clear_samplers(start=10, end=100)



import warnings
from collections import deque

from _moderngl import Attribute, Error, InvalidObject, StorageBlock, Subroutine, Uniform, UniformBlock, Varying
from _moderngl import parse_spv_inputs as _parse_spv

try:
    from moderngl import mgl
except ImportError:
    pass

__version__ = "5.9.0"


class _store:
    default_context = None


class Buffer:
    def __init__(self):
        self.mglo = None
        self._size = None
        self._dynamic = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def size(self):
        return self.mglo.size()

    @property
    def dynamic(self):
        return self._dynamic

    @property
    def glo(self):
        return self._glo

    def write(self, data, offset=0):
        self.mglo.write(data, offset)

    def write_chunks(self, data, start, step, count):
        self.mglo.write_chunks(data, start, step, count)

    def read(self, size=-1, offset=0):
        return self.mglo.read(size, offset)

    def read_into(self, buffer, size=-1, offset=0, write_offset=0):
        return self.mglo.read_into(buffer, size, offset, write_offset)

    def read_chunks(self, chunk_size, start, step, count):
        return self.mglo.read_chunks(chunk_size, start, step, count)

    def read_chunks_into(self, buffer, chunk_size, start, step, count, write_offset=0):
        return self.mglo.read(buffer, chunk_size, start, step, count, write_offset)

    def clear(self, size=-1, offset=0, chunk=None):
        self.mglo.clear(size, offset, chunk)

    def bind_to_uniform_block(self, binding=0, offset=0, size=-1):
        self.mglo.bind_to_uniform_block(binding, offset, size)

    def bind_to_storage_buffer(self, binding=0, offset=0, size=-1):
        self.mglo.bind_to_storage_buffer(binding, offset, size)

    def orphan(self, size=-1):
        self.mglo.orphan(size)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()

    def bind(self, *attribs, layout=None):
        return (self, layout, *attribs)

    def assign(self, index):
        return (self, index)


class ConditionalRender:
    def __init__(self):
        self.mglo = None
        raise TypeError()

    def __enter__(self):
        self.mglo.begin_render()
        return self

    def __exit__(self, *args):
        self.mglo.end_render()


class Query:
    def __init__(self):
        self.mglo = None
        self.crender = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __enter__(self):
        self.mglo.begin()
        return self

    def __exit__(self, *args):
        self.mglo.end()

    @property
    def samples(self):
        return self.mglo.samples

    @property
    def primitives(self):
        return self.mglo.primitives

    @property
    def elapsed(self):
        return self.mglo.elapsed


class ComputeShader:
    def __init__(self):
        self.mglo = None
        self._members = {}
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def __getitem__(self, key):
        return self._members[key]

    def __setitem__(self, key, value):
        self._members[key].value = value

    def __iter__(self):
        yield from self._members

    @property
    def glo(self):
        return self._glo

    def run(self, group_x=1, group_y=1, group_z=1):
        return self.mglo.run(group_x, group_y, group_z)

    def run_indirect(self, buffer, offset=0):
        return self.mglo.run_indirect(buffer.mglo, offset)

    def get(self, key, default):
        return self._members.get(key, default)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class Framebuffer:
    def __init__(self):
        self.mglo = None
        self._color_attachments = None
        self._depth_attachment = None
        self._size = (None, None)
        self._samples = None
        self._glo = None
        self.ctx = None
        self._is_reference = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        # Don't delete default framebuffer or a reference
        if self._is_reference:
            return

        # If object was initialized properly (ctx present) and gc_mode is auto
        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def viewport(self):
        return self.mglo.viewport

    @viewport.setter
    def viewport(self, value):
        x, y, w, h = value
        self.mglo.viewport = (int(x), int(y), int(w), int(h))

    @property
    def scissor(self):
        return self.mglo.scissor

    @scissor.setter
    def scissor(self, value):
        if value is None:
            self.mglo.scissor = None
        else:
            x, y, w, h = value
            self.mglo.scissor = (int(x), int(y), int(w), int(h))

    @property
    def color_mask(self):
        return self.mglo.color_mask

    @color_mask.setter
    def color_mask(self, value):
        self.mglo.color_mask = value

    @property
    def depth_mask(self):
        return self.mglo.depth_mask

    @depth_mask.setter
    def depth_mask(self, value):
        self.mglo.depth_mask = value

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    @property
    def size(self):
        return self._size

    @property
    def samples(self):
        return self._samples

    @property
    def bits(self):
        return self.mglo.bits

    @property
    def color_attachments(self):
        return self._color_attachments

    @property
    def depth_attachment(self):
        return self._depth_attachment

    @property
    def glo(self):
        return self._glo

    def clear(self, red=0.0, green=0.0, blue=0.0, alpha=0.0, depth=1.0, viewport=None, color=None):
        if color is not None:
            red, green, blue, alpha, *_ = tuple(color) + (0.0, 0.0, 0.0, 0.0)

        if viewport is not None:
            viewport = tuple(viewport)

        self.mglo.clear(red, green, blue, alpha, depth, viewport)

    def use(self):
        self.ctx.fbo = self
        self.mglo.use()

    def read(self, viewport=None, components=3, attachment=0, alignment=1, dtype="f1", clamp=False):
        if viewport is None:
            viewport = (0, 0, self.width, self.height)
        if len(viewport) == 2:
            viewport = (0, 0, *viewport)
        res, mem = mgl.writable_bytes(mgl.expected_size(viewport[2], viewport[3], 1, components, alignment, dtype))
        self.mglo.read_into(mem, viewport, components, attachment, alignment, clamp, dtype, 0)
        return res

    def read_into(
        self, buffer, viewport=None, components=3, attachment=0, alignment=1, dtype="f1", clamp=False, write_offset=0
    ):
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, viewport, components, attachment, alignment, clamp, dtype, write_offset)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self._color_attachments = None
            self._depth_attachment = None
            self.mglo.release()
            self.mglo = InvalidObject()


class Program:
    def __init__(self):
        self.mglo = None
        self._members = {}
        self._subroutines = None
        self._geom = (None, None, None)
        self._glo = None
        self._is_transform = None
        self._attribute_locations = None
        self._attribute_types = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def __getitem__(self, key):
        return self._members[key]

    def __setitem__(self, key, value):
        self._members[key].value = value

    def __iter__(self):
        yield from self._members

    @property
    def is_transform(self):
        return self._is_transform

    @property
    def geometry_input(self):
        return self._geom[0]

    @property
    def geometry_output(self):
        return self._geom[1]

    @property
    def geometry_vertices(self):
        return self._geom[2]

    @property
    def subroutines(self):
        return self._subroutines

    @property
    def glo(self):
        return self._glo

    def get(self, key, default):
        return self._members.get(key, default)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class Renderbuffer:
    def __init__(self):
        self.mglo = None
        self._size = (None, None)
        self._components = None
        self._samples = None
        self._depth = None
        self._dtype = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    @property
    def size(self):
        return self._size

    @property
    def samples(self):
        return self._samples

    @property
    def components(self):
        return self._components

    @property
    def depth(self):
        return self._depth

    @property
    def dtype(self):
        return self._dtype

    @property
    def glo(self):
        return self._glo

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class Sampler:
    def __init__(self):
        self.mglo = None
        self._glo = None
        self.ctx = None
        self.extra = None
        self.texture = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def use(self, location=0):
        if self.texture is not None:
            self.texture.use(location)
        self.mglo.use(location)

    def clear(self, location=0):
        self.mglo.clear(location)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()

    @property
    def repeat_x(self):
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self):
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def repeat_z(self):
        return self.mglo.repeat_z

    @repeat_z.setter
    def repeat_z(self, value):
        self.mglo.repeat_z = value

    @property
    def filter(self):
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def compare_func(self):
        return self.mglo.compare_func

    @compare_func.setter
    def compare_func(self, value):
        self.mglo.compare_func = value

    @property
    def anisotropy(self):
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self.mglo.anisotropy = value

    @property
    def border_color(self):
        return self.mglo.border_color

    @border_color.setter
    def border_color(self, value):
        self.mglo.border_color = value

    @property
    def min_lod(self):
        return self.mglo.min_lod

    @min_lod.setter
    def min_lod(self, value):
        self.mglo.min_lod = value

    @property
    def max_lod(self):
        return self.mglo.max_lod

    @max_lod.setter
    def max_lod(self, value):
        self.mglo.max_lod = value

    def assign(self, index):
        return (self, index)


class Scope:
    def __init__(self):
        self.mglo = None
        self.ctx = None
        self._framebuffer = None
        self._textures = None
        self._uniform_buffers = None
        self._storage_buffers = None
        self._samplers = None
        self.extra = None
        raise TypeError()

    def __enter__(self):
        self.mglo.begin()
        return self

    def __exit__(self, *args):
        self.mglo.end()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self._framebuffer = None
            self._textures = None
            self._uniform_buffers = None
            self._storage_buffers = None
            self._samplers = None
            self.mglo.release()
            self.mglo = InvalidObject()


class Texture:
    def __init__(self):
        self.mglo = None
        self._size = (None, None)
        self._components = None
        self._samples = None
        self._dtype = None
        self._depth = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def repeat_x(self):
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self):
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def filter(self):
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def anisotropy(self):
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self.mglo.anisotropy = value

    @property
    def swizzle(self):
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def compare_func(self):
        return self.mglo.compare_func

    @compare_func.setter
    def compare_func(self, value):
        self.mglo.compare_func = value

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    @property
    def size(self):
        return self._size

    @property
    def components(self):
        return self._components

    @property
    def samples(self):
        return self._samples

    @property
    def dtype(self):
        return self._dtype

    @property
    def depth(self):
        return self._depth

    @property
    def glo(self):
        return self._glo

    def read(self, level=0, alignment=1):
        return self.mglo.read(level, alignment)

    def read_into(self, buffer, level=0, alignment=1, write_offset=0):
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, level, alignment, write_offset)

    def write(self, data, viewport=None, level=0, alignment=1):
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(data, viewport, level, alignment)

    def build_mipmaps(self, base=0, max_level=1000):
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0):
        self.mglo.use(location)

    def bind_to_image(self, unit, read=True, write=True, level=0, format=0):
        self.mglo.bind(unit, read, write, level, format)

    def get_handle(self, resident=True):
        return self.mglo.get_handle(resident)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class Texture3D:
    def __init__(self):
        self.mglo = None
        self._size = (None, None, None)
        self._components = None
        self._samples = None
        self._dtype = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def repeat_x(self):
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self):
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def repeat_z(self):
        return self.mglo.repeat_z

    @repeat_z.setter
    def repeat_z(self, value):
        self.mglo.repeat_z = value

    @property
    def filter(self):
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def swizzle(self):
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    @property
    def depth(self):
        return self._size[2]

    @property
    def size(self):
        return self._size

    @property
    def components(self):
        return self._components

    @property
    def dtype(self):
        return self._dtype

    @property
    def glo(self):
        return self._glo

    def read(self, alignment=1):
        return self.mglo.read(alignment)

    def read_into(self, buffer, alignment=1, write_offset=0):
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(self, data, viewport=None, alignment=1):
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base=0, max_level=1000):
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0):
        self.mglo.use(location)

    def bind_to_image(self, unit, read=True, write=True, level=0, format=0):
        self.mglo.bind(unit, read, write, level, format)

    def get_handle(self, resident=True):
        return self.mglo.get_handle(resident)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class TextureCube:
    def __init__(self):
        self.mglo = None
        self._size = (None, None)
        self._components = None
        self._depth = None
        self._dtype = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def size(self):
        return self._size

    @property
    def components(self):
        return self._components

    @property
    def dtype(self):
        return self._dtype

    @property
    def filter(self):
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def swizzle(self):
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def compare_func(self):
        return self.mglo.compare_func

    @compare_func.setter
    def compare_func(self, value):
        self.mglo.compare_func = value

    @property
    def anisotropy(self):
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self.mglo.anisotropy = value

    @property
    def depth(self):
        return self._depth

    @property
    def glo(self):
        return self._glo

    def read(self, face, alignment=1):
        return self.mglo.read(face, alignment)

    def read_into(self, buffer, face, alignment=1, write_offset=0):
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, face, alignment, write_offset)

    def write(self, face, data, viewport=None, alignment=1):
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(face, data, viewport, alignment)

    def build_mipmaps(self, base=0, max_level=1000):
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0):
        self.mglo.use(location)

    def bind_to_image(self, unit, read=True, write=True, level=0, format=0):
        self.mglo.bind(unit, read, write, level, format)

    def get_handle(self, resident=True):
        return self.mglo.get_handle(resident)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class TextureArray:
    def __init__(self):
        self.mglo = None
        self._size = (None, None, None)
        self._components = None
        self._samples = None
        self._dtype = None
        self._depth = None
        self._glo = None
        self.ctx = None
        self.extra = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def repeat_x(self):
        return self.mglo.repeat_x

    @repeat_x.setter
    def repeat_x(self, value):
        self.mglo.repeat_x = value

    @property
    def repeat_y(self):
        return self.mglo.repeat_y

    @repeat_y.setter
    def repeat_y(self, value):
        self.mglo.repeat_y = value

    @property
    def filter(self):
        return self.mglo.filter

    @filter.setter
    def filter(self, value):
        self.mglo.filter = value

    @property
    def swizzle(self):
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value):
        self.mglo.swizzle = value

    @property
    def anisotropy(self):
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value):
        self.mglo.anisotropy = value

    @property
    def width(self):
        return self._size[0]

    @property
    def height(self):
        return self._size[1]

    @property
    def layers(self):
        return self._size[2]

    @property
    def size(self):
        return self._size

    @property
    def components(self):
        return self._components

    @property
    def dtype(self):
        return self._dtype

    @property
    def glo(self):
        return self._glo

    def read(self, alignment=1):
        return self.mglo.read(alignment)

    def read_into(self, buffer, alignment=1, write_offset=0):
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, alignment, write_offset)

    def write(self, data, viewport=None, alignment=1):
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(data, viewport, alignment)

    def build_mipmaps(self, base=0, max_level=1000):
        self.mglo.build_mipmaps(base, max_level)

    def use(self, location=0):
        self.mglo.use(location)

    def bind_to_image(self, unit, read=True, write=True, level=0, format=0):
        self.mglo.bind(unit, read, write, level, format)

    def get_handle(self, resident=True):
        return self.mglo.get_handle(resident)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


class VertexArray:
    def __init__(self):
        self.mglo = None
        self._program = None
        self._index_buffer = None
        self._content = None
        self._index_element_size = None
        self._glo = None
        self._mode = None
        self.ctx = None
        self.extra = None
        self.scope = None
        raise TypeError()

    def __del__(self):
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = value

    @property
    def program(self):
        return self._program

    @property
    def index_buffer(self):
        return self._index_buffer

    @property
    def index_element_size(self):
        return self._index_element_size

    @property
    def vertices(self):
        return self.mglo.vertices

    @vertices.setter
    def vertices(self, value):
        self.mglo.vertices = int(value)

    @property
    def instances(self):
        return self.mglo.instances

    @instances.setter
    def instances(self, value):
        self.mglo.instances = int(value)

    @property
    def subroutines(self):
        return self.mglo.subroutines

    @subroutines.setter
    def subroutines(self, value):
        self.mglo.subroutines = tuple(value)

    @property
    def glo(self):
        return self._glo

    def render(self, mode=None, vertices=-1, first=0, instances=-1):
        if mode is None:
            mode = self._mode

        if self.scope:
            with self.scope:
                self.mglo.render(mode, vertices, first, instances)
        else:
            self.mglo.render(mode, vertices, first, instances)

    def render_indirect(self, buffer, mode=None, count=-1, first=0):
        if mode is None:
            mode = self._mode

        if self.scope:
            with self.scope:
                self.mglo.render_indirect(buffer.mglo, mode, count, first)
        else:
            self.mglo.render_indirect(buffer.mglo, mode, count, first)

    def transform(self, buffer, mode=None, vertices=-1, first=0, instances=-1, buffer_offset=0):
        if mode is None:
            mode = self._mode

        if isinstance(buffer, (list, tuple)):
            outputs = [buf.mglo for buf in buffer]
        else:
            outputs = [buffer.mglo]

        if self.scope:
            with self.scope:
                self.mglo.transform(outputs, mode, vertices, first, instances, buffer_offset)
        else:
            self.mglo.transform(outputs, mode, vertices, first, instances, buffer_offset)

    def bind(self, attribute, cls, buffer, fmt, offset=0, stride=0, divisor=0, normalize=False):
        self.mglo.bind(attribute, cls, buffer.mglo, fmt, offset, stride, divisor, normalize)

    def release(self):
        if not isinstance(self.mglo, InvalidObject):
            self._program = None
            self._index_buffer = None
            self._content = None
            self.mglo.release()
            self.mglo = InvalidObject()


class Context:
    _valid_gc_modes = [None, "context_gc", "auto"]

    # Context Flags

    NOTHING = 0
    BLEND = 1
    DEPTH_TEST = 2
    CULL_FACE = 4
    RASTERIZER_DISCARD = 8
    PROGRAM_POINT_SIZE = 16

    # Primitive modes

    POINTS = 0x0000
    LINES = 0x0001
    LINE_LOOP = 0x0002
    LINE_STRIP = 0x0003
    TRIANGLES = 0x0004
    TRIANGLE_STRIP = 0x0005
    TRIANGLE_FAN = 0x0006
    LINES_ADJACENCY = 0x000A
    LINE_STRIP_ADJACENCY = 0x000B
    TRIANGLES_ADJACENCY = 0x000C
    TRIANGLE_STRIP_ADJACENCY = 0x0000D
    PATCHES = 0x000E

    # Texture filters

    NEAREST = 0x2600
    LINEAR = 0x2601
    NEAREST_MIPMAP_NEAREST = 0x2700
    LINEAR_MIPMAP_NEAREST = 0x2701
    NEAREST_MIPMAP_LINEAR = 0x2702
    LINEAR_MIPMAP_LINEAR = 0x2703

    # Blend function constants

    ZERO = 0x0000
    ONE = 0x0001
    SRC_COLOR = 0x0300
    ONE_MINUS_SRC_COLOR = 0x0301
    SRC_ALPHA = 0x0302
    ONE_MINUS_SRC_ALPHA = 0x0303
    DST_ALPHA = 0x0304
    ONE_MINUS_DST_ALPHA = 0x0305
    DST_COLOR = 0x0306
    ONE_MINUS_DST_COLOR = 0x0307

    DEFAULT_BLENDING = (SRC_ALPHA, ONE_MINUS_SRC_ALPHA)
    ADDITIVE_BLENDING = (ONE, ONE)
    PREMULTIPLIED_ALPHA = (SRC_ALPHA, ONE)

    # Blend equations

    FUNC_ADD = 0x8006
    FUNC_SUBTRACT = 0x800A
    FUNC_REVERSE_SUBTRACT = 0x800B
    MIN = 0x8007
    MAX = 0x8008

    # Provoking vertex

    FIRST_VERTEX_CONVENTION = 0x8E4D
    LAST_VERTEX_CONVENTION = 0x8E4E

    # Memory barrier

    VERTEX_ATTRIB_ARRAY_BARRIER_BIT = 0x00000001
    ELEMENT_ARRAY_BARRIER_BIT = 0x00000002
    UNIFORM_BARRIER_BIT = 0x00000004
    TEXTURE_FETCH_BARRIER_BIT = 0x00000008
    SHADER_IMAGE_ACCESS_BARRIER_BIT = 0x00000020
    COMMAND_BARRIER_BIT = 0x00000040
    PIXEL_BUFFER_BARRIER_BIT = 0x00000080
    TEXTURE_UPDATE_BARRIER_BIT = 0x00000100
    BUFFER_UPDATE_BARRIER_BIT = 0x00000200
    FRAMEBUFFER_BARRIER_BIT = 0x00000400
    TRANSFORM_FEEDBACK_BARRIER_BIT = 0x00000800
    ATOMIC_COUNTER_BARRIER_BIT = 0x00001000
    SHADER_STORAGE_BARRIER_BIT = 0x00002000
    ALL_BARRIER_BITS = 0xFFFFFFFF

    def __init__(self):
        self.mglo = None
        self._screen = None
        self._info = None
        self._extensions = None
        self.version_code = None
        self.fbo = None
        self.extra = None
        self._gc_mode = None
        self._objects = deque()
        raise TypeError()

    def __del__(self):
        if hasattr(self, "_gc_mode") and self._gc_mode == "auto":
            self.release()

    @property
    def gc_mode(self):
        return self._gc_mode

    @gc_mode.setter
    def gc_mode(self, value):
        if value not in self._valid_gc_modes:
            raise ValueError("Valid  gc modes:", self._valid_gc_modes)

        self._gc_mode = value

    @property
    def objects(self):
        return self._objects

    def gc(self):
        count = 0
        # Keep iterating until there are no more objects.
        # An object deletion can trigger new objects to be added
        while self._objects:
            # Remove the oldest objects first
            obj = self._objects.popleft()
            obj.release()
            count += 1

        return count

    @property
    def line_width(self):
        return self.mglo.line_width

    @line_width.setter
    def line_width(self, value):
        self.mglo.line_width = value

    @property
    def point_size(self):
        return self.mglo.point_size

    @point_size.setter
    def point_size(self, value):
        self.mglo.point_size = value

    @property
    def depth_func(self):
        raise NotImplementedError()

    @depth_func.setter
    def depth_func(self, value):
        self.mglo.depth_func = value

    @property
    def blend_func(self):
        raise NotImplementedError()

    @blend_func.setter
    def blend_func(self, value):
        self.mglo.blend_func = tuple(value)

    @property
    def blend_equation(self):
        raise NotImplementedError()

    @blend_equation.setter
    def blend_equation(self, value):
        if not isinstance(value, tuple):
            self.mglo.blend_equation = tuple([value])
        else:
            self.mglo.blend_equation = tuple(value)

    @property
    def depth_clamp_range(self):
        raise NotImplementedError()

    @depth_clamp_range.setter
    def depth_clamp_range(self, value):
        self.mglo.depth_clamp_range = value

    @property
    def multisample(self):
        raise NotImplementedError()

    @multisample.setter
    def multisample(self, value):
        self.mglo.multisample = value

    @property
    def provoking_vertex(self):
        raise NotImplementedError()

    @provoking_vertex.setter
    def provoking_vertex(self, value):
        self.mglo.provoking_vertex = value

    @property
    def polygon_offset(self):
        return self.mglo.polygon_offset

    @polygon_offset.setter
    def polygon_offset(self, value):
        factor, units = value
        self.mglo.polygon_offset = (float(factor), float(units))

    @property
    def viewport(self):
        return self.mglo.fbo.viewport

    @viewport.setter
    def viewport(self, value):
        x, y, w, h = value
        self.mglo.fbo.viewport = (int(x), int(y), int(w), int(h))

    @property
    def scissor(self):
        return self.mglo.fbo.scissor

    @scissor.setter
    def scissor(self, value):
        if value is None:
            self.mglo.fbo.scissor = None
        else:
            x, y, w, h = value
            self.mglo.fbo.scissor = (int(x), int(y), int(w), int(h))

    @property
    def max_samples(self):
        return self.mglo.max_samples

    @property
    def max_integer_samples(self):
        return self.mglo.max_integer_samples

    @property
    def max_texture_units(self):
        return self.mglo.max_texture_units

    @property
    def default_texture_unit(self):
        return self.mglo.default_texture_unit

    @default_texture_unit.setter
    def default_texture_unit(self, value):
        self.mglo.default_texture_unit = value

    @property
    def max_anisotropy(self):
        return self.mglo.max_anisotropy

    @property
    def screen(self):
        return self._screen

    @property
    def wireframe(self):
        return self.mglo.wireframe

    @wireframe.setter
    def wireframe(self, value):
        self.mglo.wireframe = value

    @property
    def front_face(self):
        return self.mglo.front_face

    @front_face.setter
    def front_face(self, value):
        self.mglo.front_face = str(value)

    @property
    def cull_face(self):
        return self.mglo.cull_face

    @cull_face.setter
    def cull_face(self, value):
        self.mglo.cull_face = str(value)

    @property
    def patch_vertices(self):
        return self.mglo.patch_vertices

    @patch_vertices.setter
    def patch_vertices(self, value):
        self.mglo.patch_vertices = value

    @property
    def error(self):
        return self.mglo.error

    @property
    def extensions(self):
        if self._extensions is None:
            self._extensions = self.mglo.extensions

        return self._extensions

    @property
    def info(self):
        if self._info is None:
            self._info = self.mglo.info

        return self._info

    @property
    def includes(self):
        return self.mglo.includes

    def clear(self, red=0.0, green=0.0, blue=0.0, alpha=0.0, depth=1.0, viewport=None, color=None):
        if color is not None:
            red, green, blue, alpha, *_ = tuple(color) + (0.0, 0.0, 0.0, 0.0)

        self.mglo.fbo.clear(red, green, blue, alpha, depth, viewport)

    def enable_only(self, flags):
        self.mglo.enable_only(flags)

    def enable(self, flags):
        self.mglo.enable(flags)

    def disable(self, flags):
        self.mglo.disable(flags)

    def enable_direct(self, enum):
        self.mglo.enable_direct(enum)

    def disable_direct(self, enum):
        self.mglo.disable_direct(enum)

    def finish(self):
        self.mglo.finish()

    def copy_buffer(self, dst: Buffer, src: Buffer, size=-1, read_offset=0, write_offset=0):
        self.mglo.copy_buffer(dst.mglo, src.mglo, size, read_offset, write_offset)

    def copy_framebuffer(self, dst, src):
        self.mglo.copy_framebuffer(dst.mglo, src.mglo)

    def detect_framebuffer(self, glo=None):
        res = Framebuffer.__new__(Framebuffer)
        res.mglo, res._size, res._samples, res._glo = self.mglo.detect_framebuffer(glo)
        res._color_attachments = None
        res._depth_attachment = None
        res.ctx = self
        res._is_reference = True
        res.extra = None
        return res

    def buffer(self, data=None, reserve=0, dynamic=False):
        if type(reserve) is str:
            reserve = mgl.strsize(reserve)

        res = Buffer.__new__(Buffer)
        res.mglo, res._size, res._glo = self.mglo.buffer(data, reserve, dynamic)
        res._dynamic = dynamic
        res.ctx = self
        res.extra = None
        return res

    def external_texture(self, glo, size, components, samples, dtype):
        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.external_texture(glo, size, components, samples, dtype)
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def texture(
        self,
        size,
        components,
        data=None,
        samples=0,
        alignment=1,
        dtype="f1",
        internal_format=None,
        renderbuffer=False,
    ):
        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.texture(
            size, components, data, samples, alignment, dtype, internal_format or 0, renderbuffer
        )
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def texture_array(self, size, components, data=None, alignment=1, dtype="f1"):
        res = TextureArray.__new__(TextureArray)
        res.mglo, res._glo = self.mglo.texture_array(size, components, data, alignment, dtype)
        res._size = size
        res._components = components
        res._dtype = dtype
        res.ctx = self
        res.extra = None
        return res

    def texture3d(self, size, components, data=None, alignment=1, dtype="f1"):
        res = Texture3D.__new__(Texture3D)
        res._size = size
        res._components = components
        res._dtype = dtype
        res.mglo, res._glo = self.mglo.texture3d(size, components, data, alignment, dtype)
        res.ctx = self
        res.extra = None
        return res

    def texture_cube(self, size, components, data=None, alignment=1, dtype="f1", internal_format=None):
        res = TextureCube.__new__(TextureCube)
        res.mglo, res._glo = self.mglo.texture_cube(size, components, data, alignment, dtype, internal_format or 0)
        res._size = size
        res._components = components
        res._dtype = dtype
        res.ctx = self
        res.extra = None
        return res

    def depth_texture(self, size, data=None, samples=0, alignment=4, renderbuffer=False):
        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.depth_texture(size, data, samples, alignment, renderbuffer)
        res._size = size
        res._components = 1
        res._samples = samples
        res._dtype = "f4"
        res._depth = True
        res.ctx = self
        res.extra = None
        return res

    def depth_texture_cube(self, size, data=None, alignment=4):
        res = TextureCube.__new__(TextureCube)
        res.mglo, res._glo = self.mglo.depth_texture_cube(size, data, alignment)
        res._size = size
        res._components = 1
        res._dtype = "f4"
        res._depth = True
        res.ctx = self
        res.extra = None
        return res

    def vertex_array(self, *args, **kwargs):
        if len(args) > 2 and type(args[1]) is Buffer:
            return self.simple_vertex_array(*args, **kwargs)
        return self._vertex_array(*args, **kwargs)

    def _vertex_array(self, program, content, index_buffer=None, index_element_size=4, skip_errors=False, mode=None):
        locations = program._attribute_locations
        types = program._attribute_types
        index_buffer_mglo = None if index_buffer is None else index_buffer.mglo
        mgl_content = []

        for buffer, layout, *attribs in content:
            if layout is None:
                layout = detect_format(program, attribs)
            attribs = [types[x] if type(x) is int else types[locations[x]] for x in attribs]
            mgl_content.append((buffer.mglo, layout, *attribs))

        res = VertexArray.__new__(VertexArray)
        res.mglo, res._glo = self.mglo.vertex_array(
            program.mglo,
            tuple(mgl_content),
            index_buffer_mglo,
            index_element_size,
            skip_errors,
        )
        res._program = program
        res._index_buffer = index_buffer
        res._content = content
        res._index_element_size = index_element_size
        if mode is not None:
            res._mode = mode
        else:
            res._mode = self.POINTS if program.is_transform else self.TRIANGLES
        res.ctx = self
        res.extra = None
        res.scope = None
        return res

    def simple_vertex_array(self, program, buffer, *attributes, index_buffer=None, index_element_size=4, mode=None):
        if type(buffer) is list:
            raise SyntaxError("Change simple_vertex_array to vertex_array")

        content = [(buffer, detect_format(program, attributes)) + attributes]
        return self._vertex_array(program, content, index_buffer, index_element_size, mode=mode)

    def program(
        self,
        vertex_shader=None,
        fragment_shader=None,
        geometry_shader=None,
        tess_control_shader=None,
        tess_evaluation_shader=None,
        varyings=(),
        fragment_outputs=None,
        attributes=None,
        varyings_capture_mode="interleaved",
    ):
        if varyings_capture_mode not in ("interleaved", "separate"):
            raise ValueError("varyings_capture_mode must be interleaved or separate")

        if type(varyings) is str:
            varyings = (varyings,)

        varyings = tuple(varyings)

        if fragment_outputs is None:
            fragment_outputs = {}

        res = Program.__new__(Program)
        res.mglo, _members, res._subroutines, res._geom, res._glo = self.mglo.program(
            vertex_shader,
            fragment_shader,
            geometry_shader,
            tess_control_shader,
            tess_evaluation_shader,
            None,
            varyings,
            fragment_outputs,
            varyings_capture_mode == "interleaved",
        )
        res._members, res._attribute_locations, res._attribute_types = _members

        if isinstance(vertex_shader, bytes) and int.from_bytes(vertex_shader[:4], "little") == 0x07230203:
            res._attribute_types = _parse_spv(res._glo, vertex_shader)
            for info in res._attribute_types.values():
                res._attribute_locations[info.name] = info.location

        if attributes is not None:
            res._attribute_locations = {}
            for i, name in enumerate(attributes):
                res._attribute_locations[name] = i

        res._is_transform = fragment_shader is None
        res.ctx = self
        res.extra = None
        return res

    def query(self, samples=False, any_samples=False, time=False, primitives=False):
        res = Query.__new__(Query)
        res.mglo = self.mglo.query(samples, any_samples, time, primitives)
        res.crender = None

        if samples or any_samples:
            res.crender = ConditionalRender.__new__(ConditionalRender)
            res.crender.mglo = res.mglo

        res.ctx = self
        res.extra = None
        return res

    def scope(
        self,
        framebuffer=None,
        enable_only=None,
        textures=(),
        uniform_buffers=(),
        storage_buffers=(),
        samplers=(),
        enable=None,
    ):
        if enable is not None:
            enable_only = enable

        if framebuffer is None:
            framebuffer = self.screen
            if framebuffer is None:
                raise RuntimeError("A framebuffer must be specified")

        mgl_textures = tuple((tex.mglo, idx) for tex, idx in textures)
        mgl_uniform_buffers = tuple((buf.mglo, idx) for buf, idx in uniform_buffers)
        mgl_storage_buffers = tuple((buf.mglo, idx) for buf, idx in storage_buffers)

        res = Scope.__new__(Scope)
        res.mglo = self.mglo.scope(
            framebuffer.mglo,
            enable_only,
            mgl_textures,
            mgl_uniform_buffers,
            mgl_storage_buffers,
            samplers,
        )
        res.ctx = self
        res._framebuffer = framebuffer
        res._textures = textures
        res._uniform_buffers = uniform_buffers
        res._storage_buffers = storage_buffers
        res._samplers = samplers
        res.extra = None
        return res

    def simple_framebuffer(self, size, components=4, samples=0, dtype="f1"):
        return self.framebuffer(
            self.renderbuffer(size, components, samples=samples, dtype=dtype),
            self.depth_renderbuffer(size, samples=samples),
        )

    def framebuffer(self, color_attachments=(), depth_attachment=None):
        if type(color_attachments) is Texture or type(color_attachments) is Renderbuffer:
            color_attachments = (color_attachments,)

        ca_mglo = tuple(x.mglo for x in color_attachments)
        da_mglo = None if depth_attachment is None else depth_attachment.mglo

        res = Framebuffer.__new__(Framebuffer)
        res.mglo, res._size, res._samples, res._glo = self.mglo.framebuffer(ca_mglo, da_mglo)
        res._color_attachments = tuple(color_attachments)
        res._depth_attachment = depth_attachment
        res.ctx = self
        res._is_reference = False
        res.extra = None
        return res

    def empty_framebuffer(self, size, layers=0, samples=0):
        res = Framebuffer.__new__(Framebuffer)
        res.mglo, res._size, res._samples, res._glo = self.mglo.empty_framebuffer(size, layers, samples)
        res._color_attachments = ()
        res._depth_attachment = None
        res.ctx = self
        res._is_reference = False
        res.extra = None
        return res

    def renderbuffer(self, size, components=4, samples=0, dtype="f1"):
        res = Renderbuffer.__new__(Renderbuffer)
        res.mglo, res._glo = self.mglo.texture(size, components, None, samples, 1, dtype, 0, True)
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def depth_renderbuffer(self, size, samples=0):
        res = Renderbuffer.__new__(Renderbuffer)
        res.mglo, res._glo = self.mglo.depth_texture(size, None, samples, 1, True)
        res._size = size
        res._components = 1
        res._samples = samples
        res._dtype = "f4"
        res._depth = True
        res.ctx = self
        res.extra = None
        return res

    def compute_shader(self, source):
        res = ComputeShader.__new__(ComputeShader)
        res.mglo, _members, _, _, res._glo = self.mglo.program(
            None,
            None,
            None,
            None,
            None,
            source,
            (),
            {},
            False,
        )
        res._members = _members[0]

        res.ctx = self
        res.extra = None
        return res

    def sampler(
        self,
        repeat_x=True,
        repeat_y=True,
        repeat_z=True,
        filter=None,
        anisotropy=1.0,
        compare_func="?",
        border_color=None,
        min_lod=-1000.0,
        max_lod=1000.0,
        texture=None,
    ):
        res = Sampler.__new__(Sampler)
        res.mglo, res._glo = self.mglo.sampler()
        res.ctx = self
        res.repeat_x = repeat_x
        res.repeat_y = repeat_y
        res.repeat_z = repeat_z
        res.filter = filter or (9729, 9729)
        res.anisotropy = anisotropy
        res.compare_func = compare_func
        if border_color:
            res.border_color = border_color
        res.min_lod = min_lod
        res.max_lod = max_lod
        res.extra = None
        res.texture = texture
        return res

    def memory_barrier(self, barriers=ALL_BARRIER_BITS, by_region=False):
        self.mglo.memory_barrier(barriers, by_region)

    def clear_samplers(self, start=0, end=-1):
        self.mglo.clear_samplers(start, end)

    def core_profile_check(self):
        profile_mask = self.info["GL_CONTEXT_PROFILE_MASK"]
        if profile_mask != 1:
            warnings.warn("The window should request a CORE OpenGL profile")

        version_code = self.version_code
        if not version_code:
            major, minor = map(int, self.info["GL_VERSION"].split(".", 2)[:2])
            version_code = major * 100 + minor * 10

        if version_code < 330:
            warnings.warn("The window should support OpenGL 3.3+ (version_code=%d)" % version_code)

    def __enter__(self):
        self.mglo.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mglo.__exit__(exc_type, exc_val, exc_tb)

    def release(self):
        if _store.default_context is self:
            _store.default_context = None
        if not isinstance(self.mglo, InvalidObject):
            self.mglo.release()
            self.mglo = InvalidObject()


def create_context(require=None, standalone=False, share=False, **settings):
    if require is None:
        require = 330

    mode = "standalone" if standalone is True else "detect"
    if share is True:
        mode = "share"

    ctx = Context.__new__(Context)
    ctx.mglo, ctx.version_code = mgl.create_context(glversion=require, mode=mode, **settings)
    ctx._info = None
    ctx._extensions = None
    ctx.extra = None
    ctx._gc_mode = None
    ctx._objects = deque()

    if ctx.version_code < require:
        raise ValueError("Requested OpenGL version {0}, got version {1}".format(require, ctx.version_code))

    if standalone:
        ctx._screen = None
        ctx.fbo = None
    else:
        # Default framebuffer
        ctx._screen = ctx.detect_framebuffer(0)
        # Currently bound framebuffer
        ctx.fbo = ctx.detect_framebuffer()
        ctx.mglo.fbo = ctx.fbo.mglo

    _store.default_context = ctx
    return ctx


def get_context():
    return _store.default_context


def create_standalone_context(**kwargs):
    return create_context(standalone=True, **kwargs)


def detect_format(program, attributes, mode="mgl"):
    def fmt(attr):
        # Translate shape format into attribute format
        mgl_fmt = {"d": "f8", "I": "u"}
        # moderngl attribute format uses f, i and u
        if mode == "mgl":
            return (
                attr.array_length * attr.dimension,
                mgl_fmt.get(attr.shape) or attr.shape,
            )
        # struct attribute format uses f, d, i and I
        elif mode == "struct":
            return attr.array_length * attr.dimension, attr.shape
        else:
            raise ValueError("invalid format mode: {0}".format(mode))

    locations = program._attribute_locations
    types = program._attribute_types
    return " ".join("%d%s" % fmt(types[x] if type(x) is int else types[locations[x]]) for x in attributes)


def _resolve_module_constants(scope):
    _constants = [
        "NOTHING",
        "BLEND",
        "DEPTH_TEST",
        "CULL_FACE",
        "RASTERIZER_DISCARD",
        "PROGRAM_POINT_SIZE",
        "POINTS",
        "LINES",
        "LINE_LOOP",
        "LINE_STRIP",
        "TRIANGLES",
        "TRIANGLE_STRIP",
        "TRIANGLE_FAN",
        "LINES_ADJACENCY",
        "LINE_STRIP_ADJACENCY",
        "TRIANGLES_ADJACENCY",
        "TRIANGLE_STRIP_ADJACENCY",
        "PATCHES",
        "NEAREST",
        "LINEAR",
        "NEAREST_MIPMAP_NEAREST",
        "LINEAR_MIPMAP_NEAREST",
        "NEAREST_MIPMAP_LINEAR",
        "LINEAR_MIPMAP_LINEAR",
        "ZERO",
        "ONE",
        "SRC_COLOR",
        "ONE_MINUS_SRC_COLOR",
        "SRC_ALPHA",
        "ONE_MINUS_SRC_ALPHA",
        "DST_ALPHA",
        "ONE_MINUS_DST_ALPHA",
        "DST_COLOR",
        "ONE_MINUS_DST_COLOR",
        "DEFAULT_BLENDING",
        "ADDITIVE_BLENDING",
        "PREMULTIPLIED_ALPHA",
        "FUNC_ADD",
        "FUNC_SUBTRACT",
        "FUNC_REVERSE_SUBTRACT",
        "MIN",
        "MAX",
        "FIRST_VERTEX_CONVENTION",
        "LAST_VERTEX_CONVENTION",
        "VERTEX_ATTRIB_ARRAY_BARRIER_BIT",
        "ELEMENT_ARRAY_BARRIER_BIT",
        "UNIFORM_BARRIER_BIT",
        "TEXTURE_FETCH_BARRIER_BIT",
        "SHADER_IMAGE_ACCESS_BARRIER_BIT",
        "COMMAND_BARRIER_BIT",
        "PIXEL_BUFFER_BARRIER_BIT",
        "TEXTURE_UPDATE_BARRIER_BIT",
        "BUFFER_UPDATE_BARRIER_BIT",
        "FRAMEBUFFER_BARRIER_BIT",
        "TRANSFORM_FEEDBACK_BARRIER_BIT",
        "ATOMIC_COUNTER_BARRIER_BIT",
        "SHADER_STORAGE_BARRIER_BIT",
        "ALL_BARRIER_BITS",
    ]

    for c in _constants:
        scope[c] = getattr(Context, c)


_resolve_module_constants(locals())



"""
Test buffer read_into method
"""


def test_1(ctx):
    buf = ctx.buffer(b'abc')
    data = bytearray(3)
    buf.read_into(data)
    assert bytes(data) == b'abc'


def test_2(ctx):
    buf = ctx.buffer(b'abcxyz123')
    data = bytearray(3)
    buf.read_into(data, offset=6)
    assert bytes(data) == b'123'


def test_3(ctx):
    buf = ctx.buffer(b'abcxyz123')
    data = bytearray(12)
    buf.read_into(data, 3, offset=3, write_offset=0)
    buf.read_into(data, 3, offset=0, write_offset=3)
    buf.read_into(data, 3, offset=6, write_offset=6)
    buf.read_into(data, 3, offset=3, write_offset=9)
    assert bytes(data) == b'xyzabc123xyz'



import moderngl
import pytest


def test_0(ctx):
    texture = ctx.texture_cube((16, 16), 4)
    assert texture.size == (16, 16)
    assert texture.components == 4
    assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert texture.swizzle == "RGBA"
    assert texture.glo > 0
    assert texture.dtype == "f1"
    assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)
    assert texture.swizzle == "RGBA"
    assert texture == texture
    assert texture.anisotropy == 0.0

    texture.anisotropy = ctx.max_anisotropy
    assert texture.anisotropy == ctx.max_anisotropy

    texture.filter = moderngl.NEAREST, moderngl.NEAREST
    assert texture.filter == (moderngl.NEAREST, moderngl.NEAREST)
    texture.swizzle = "BGRA"
    assert texture.swizzle == "BGRA"
    texture.anisotropy = ctx.max_anisotropy
    assert texture.anisotropy == ctx.max_anisotropy


def test_1(ctx):
    faces = [
        b'\x00\x00\xff' * 4 * 4,
        b'\x00\xff\x00' * 4 * 4,
        b'\x00\xff\xff' * 4 * 4,
        b'\xff\x00\x00' * 4 * 4,
        b'\xff\x00\xff' * 4 * 4,
        b'\x00\xff\x00' * 4 * 4,
    ]
    tex = ctx.texture_cube((4, 4), 3, b''.join(faces))
    assert tex.read(0) == faces[0]
    assert tex.read(1) == faces[1]
    assert tex.read(2) == faces[2]
    assert tex.read(3) == faces[3]
    assert tex.read(4) == faces[4]
    assert tex.read(5) == faces[5]
    tex.write(0, b'\xff\xff\xff' * 4 * 4)
    assert tex.read(0) == b'\xff\xff\xff' * 4 * 4
    tex.write(2, b'\xff\xff\xff' * 4 * 4)
    assert tex.read(2) == b'\xff\xff\xff' * 4 * 4
    tex.write(5, b'\xff\xff\xff' * 4 * 4)
    assert tex.read(5) == b'\xff\xff\xff' * 4 * 4


def test_2(ctx):
    tex = ctx.texture_cube((4, 4), 3)

    with pytest.raises(Exception):
        tex.write(0, b'\xff\xff\xff' * 4 * 3)

    with pytest.raises(Exception):
        tex.write(0, b'\xff\xff\xff' * 4 * 5)

    with pytest.raises(Exception):
        tex.write(-1, b'\xff\xff\xff' * 4 * 4)

    with pytest.raises(Exception):
        tex.write(6, b'\xff\xff\xff' * 4 * 4)


def test_3(ctx):
    tex = ctx.texture_cube((4, 4), 3)

    with pytest.raises(Exception):
        tex.read(-1)

    with pytest.raises(Exception):
        tex.read(6)


def test_texture_default_filter(ctx):
    """Ensure default filter is correct"""
    # Float types
    for dtype in ["f1", "f2", "f4"]:
        texture = ctx.texture_cube((10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.LINEAR, moderngl.LINEAR)

    for dtype in ["u1", "u2", "u4", "i1", "i2", "i4"]:
        texture = ctx.texture_cube((10, 10), 4, dtype=dtype)
        assert texture.filter == (moderngl.NEAREST, moderngl.NEAREST)





def test_init_from_texture_and_params(ctx):
    params_tex = dict(
        size=(5, 10),
        components=4,
        samples=0,
        dtype="f1"
    )
    tex = ctx.texture(params_tex["size"], params_tex["components"])
    ext = ctx.external_texture(**{**params_tex, **dict(glo=tex.glo)})
    assert ext.size == params_tex["size"]
    assert ext.width == params_tex["size"][0]
    assert ext.height == params_tex["size"][1]
    assert ext.components == params_tex["components"]
    assert ext.glo == tex.glo
    assert ext.dtype == params_tex["dtype"]
    assert ext.samples == params_tex["samples"]
    assert ext.repeat_x is True
    assert ext.repeat_y is True
    assert ext.filter == (ctx.LINEAR, ctx.LINEAR)
    assert ext.swizzle == "RGBA"
    assert ext.depth is False
    assert ext.anisotropy == 0.0
    assert ext.mglo != tex.mglo
    assert ext.ctx == ctx
    assert ext.extra == None



import moderngl
import pytest
import numpy as np


def test_1(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec4 in_vert;
            out vec4 out_vert;

            void main() {
                out_vert = in_vert;
            }
        ''',
        varyings=['out_vert']
    )

    vertices = [
        4.0, 2.0, 7.5, 1.8,
        3.0, 2.8, 6.0, 10.0
    ]
    count = 10
    indices = [0, 1] * 10

    # UNSIGNED_INT index
    vbo1 = ctx.buffer(np.array(vertices, dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=vbo1.size * count)
    index = ctx.buffer(np.array(indices, dtype='u4').tobytes())
    vao = ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=4)
    vao.transform(vbo2, moderngl.POINTS)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    np.testing.assert_almost_equal(res, vertices * count)

    # UNSIGNED_SHORT index
    vbo1 = ctx.buffer(np.array(vertices, dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=vbo1.size * count)
    index = ctx.buffer(np.array(indices, dtype='u2').tobytes())
    vao = ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=2)
    vao.transform(vbo2, moderngl.POINTS)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    np.testing.assert_almost_equal(res, vertices * count)

    # UNSIGNED_BYTE index
    vbo1 = ctx.buffer(np.array(vertices, dtype='f4').tobytes())
    vbo2 = ctx.buffer(reserve=vbo1.size * count)
    index = ctx.buffer(np.array(indices, dtype='u1').tobytes())
    vao = ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index, index_element_size=1)
    vao.transform(vbo2, moderngl.POINTS)
    res = np.frombuffer(vbo2.read(), dtype='f4')
    np.testing.assert_almost_equal(res, vertices * count)


def test_2(ctx):
    prog = ctx.program(
        vertex_shader='''
            #version 330

            in vec4 in_vert;
            out vec4 out_vert;

            void main() {
                out_vert = in_vert;
            }
        ''',
        varyings=['out_vert']
    )

    vertices = [
        4.0, 2.0, 7.5, 1.8,
        3.0, 2.8, 6.0, 10.0
    ]
    indices = [0, 1, 0, 1, 0, 1, 0, 1, 0]

    vbo1 = ctx.buffer(np.array(vertices, dtype='f4').tobytes())
    index_u4 = ctx.buffer(np.array(indices, dtype='u4').tobytes())

    with pytest.raises(moderngl.Error):
        ctx.simple_vertex_array(prog, vbo1, 'in_vert', index_buffer=index_u4, index_element_size=0)




def test_1(ctx):
    buf = ctx.buffer(b'abc')
    assert buf.read() == b'abc'
    buf.write(b'xyz')
    assert buf.read() == b'xyz'
    assert buf.read(offset=1) == b'yz'


def test_2(ctx):
    buf = ctx.buffer(reserve=10)
    assert buf.size == 10
    buf.write(b'12345')
    buf.write(b'abcde', offset=5)
    assert buf.read(5) == b'12345'
    assert buf.read(5, offset=5) == b'abcde'


def test_3(ctx):
    buf = ctx.buffer(reserve='1KB')
    assert buf.size == 1024
    buf.clear(chunk=b'u')
    assert buf.read() == b'u' * 1024
    buf.clear(size=10, offset=10, chunk=b'ab')
    assert buf.read(10, offset=0) == b'uuuuuuuuuu'
    assert buf.read(10, offset=10) == b'ababababab'
    assert buf.read(10, offset=20) == b'uuuuuuuuuu'


def test_4(ctx):
    buf = ctx.buffer(reserve=10)
    buf.write_chunks(b'00000', 0, 2, 5)
    buf.write_chunks(b'11111', 1, 2, 5)
    assert buf.read(10, offset=0) == b'0101010101'


def test_5(ctx):
    buf = ctx.buffer(b'123456789')
    buf.write_chunks(b'AABBCC', 0, 3, 3)
    assert buf.read() == b'AA3BB6CC9'
    assert buf.read_chunks(2, 0, 3, 3) == b'AABBCC'


def test_6(ctx):
    buf = ctx.buffer(b'123456789')
    buf.write_chunks(b'XYZ', -1, -3, 3)
    assert buf.read() == b'12Z45Y78X'
    assert buf.read_chunks(1, -1, -3, 3) == b'XYZ'


def test_7(ctx):
    buf = ctx.buffer(b'123456789')
    assert buf.read_chunks(3, 0, 3, 3) == b'123456789'


def test_8(ctx):
    buf1 = ctx.buffer(b'abc', dynamic=True)
    buf2 = ctx.buffer(b'abc', dynamic=False)
    assert buf1.dynamic is True
    assert buf2.dynamic is False



import moderngl
import pytest
import numpy as np


@pytest.fixture(scope='module')
def prog(ctx_static):
    return ctx_static.program(
        vertex_shader='''
            #version 330

            in vec2 in_vert;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
        ''',
        fragment_shader='''
            #version 330

            out vec4 fragColor;
            uniform vec4 color;

            void main() {
                fragColor = color;
            }
        ''',
    )


@pytest.fixture(scope='module')
def vao(ctx_static, prog):
    quad = [
        -1.0, 1.0,
        -1.0, -1.0,
        1.0, 1.0,
        1.0, -1.0,
    ]
    vbo = ctx_static.buffer(np.array(quad, dtype='f4'))
    return ctx_static.simple_vertex_array(prog, vbo, 'in_vert')


def create_fbo(ctx, size, components=3):
    fbo = ctx.framebuffer(
        color_attachments=[ctx.texture(size, components)],
    )
    fbo.clear()
    return fbo


def test_default_value(ctx):
    """Ensure default value of the scissor matches the framebuffer size"""
    size = (16, 32)
    fbo = create_fbo(ctx, size)
    assert fbo.scissor == (0, 0, *size)


def test_reset_scissor(ctx):
    """We should be able to reset scissor with `None`"""
    size = (16, 32)
    fbo = create_fbo(ctx, size)
    new_val = 2, 2, 14, 30

    fbo.scissor = new_val
    assert fbo.scissor == new_val

    fbo.scissor = None
    assert fbo.scissor == (0, 0, *size)


def test_render(ctx, prog, vao):
    """Render different color to the four corners of the fbo using scissor"""
    size = (4, 4)
    fbo = create_fbo(ctx, size)
    fbo.use()

    # upper left (red)
    fbo.scissor = 0, 2, 2, 2
    prog['color'].value = 1.0, 0.0, 0.0, 0.0
    vao.render(mode=moderngl.TRIANGLE_STRIP)

    # upper right (green)
    fbo.scissor = 2, 2, 2, 2
    prog['color'].value = 0.0, 1.0, 0.0, 0.0
    vao.render(mode=moderngl.TRIANGLE_STRIP)

    # lower left (green)
    fbo.scissor = 0, 0, 2, 2
    prog['color'].value = 0.0, 0.0, 1.0, 0.0
    vao.render(mode=moderngl.TRIANGLE_STRIP)

    # lower right (green)
    fbo.scissor = 2, 0, 2, 2
    prog['color'].value = 1.0, 1.0, 1.0, 1.0
    vao.render(mode=moderngl.TRIANGLE_STRIP)

    data = fbo.read()
    # Note that the data is flipped horizontally
    expected = (
        b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
        b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
        b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
        b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
    )
    assert data == expected


def test_clear(ctx):
    """Ensure fbo clearing works with scissoring"""
    size = (4, 4)
    fbo = create_fbo(ctx, size)
    fbo.use()

    # upper left (red)
    fbo.scissor = 0, 2, 2, 2
    fbo.clear(1.0, 0.0, 0.0, 0.0)

    # upper right (green)
    fbo.scissor = 2, 2, 2, 2
    fbo.clear(0.0, 1.0, 0.0, 0.0)

    # lower left (green)
    fbo.scissor = 0, 0, 2, 2
    fbo.clear(0.0, 0.0, 1.0, 0.0)

    # lower right (green)
    fbo.scissor = 2, 0, 2, 2
    fbo.clear(1.0, 1.0, 1.0, 1.0)

    data = fbo.read()
    # Note that the data is flipped horizontally
    expected = (
        b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
        b'\x00\x00\xff'b'\x00\x00\xff'b'\xff\xff\xff'b'\xff\xff\xff'
        b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
        b'\xff\x00\x00'b'\xff\x00\x00'b'\x00\xff\x00'b'\x00\xff\x00'
    )
    assert data == expected


def test_scissor_leak(ctx):
    """Make sure we don't leak scissor values to other framebuffers"""
    size = 2, 2
    fbo1 = create_fbo(ctx, size, components=4)
    fbo2 = create_fbo(ctx, size, components=4)

    fbo1.scissor = 0, 0, 1, 1
    fbo2.scissor = 1, 1, 1, 1

    fbo1.clear(color=(1.0, 0.0, 0.0, 0.0))
    fbo2.clear(color=(0.0, 1.0, 0.0, 0.0))

    assert fbo1.read() == b'\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    assert fbo2.read() == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00'


def test_clear_with_viewport(ctx):
    """Clearing with viewport take presence over scissor"""
    size = 2, 1
    fbo = create_fbo(ctx, size)
    fbo.scissor = 1, 0, 1, 1
    fbo.clear(color=(1.0, 1.0, 1.0, 1.0), viewport=(0, 0, 1, 1))
    assert fbo.read() == b'\xff\xff\xff\x00\x00\x00'



import moderngl


def test_program(ctx):
    program = ctx.program(
        vertex_shader='''
            #version 330

            uniform vec2 pos;
            uniform float scale;

            in vec2 vert;
            out vec2 v_vert;

            void main() {
                gl_Position = vec4(pos + vert * scale, 0.0, 1.0);
                v_vert = vert;
            }
        ''',
        fragment_shader='''
            #version 330

            in vec2 v_vert;
            out vec4 color;

            void main() {
                color = vec4(v_vert, 0.0, 1.0);
            }
        ''',
    )
    assert program.glo > 0

    assert 'vert' in program
    assert 'pos' in program
    assert 'scale' in program
    assert program.is_transform is False

    assert isinstance(program['vert'], moderngl.Attribute)
    assert isinstance(program['pos'], moderngl.Uniform)
    assert isinstance(program['scale'], moderngl.Uniform)


def test_geo_input_output_primitive(ctx):
    """Test detection of all geo shader input/output primitives types"""
    input_types = [
        ("points", moderngl.POINTS),
        ("lines", moderngl.LINES),
        ("lines_adjacency", moderngl.LINES_ADJACENCY),
        ("triangles", moderngl.TRIANGLES),
        ("triangles_adjacency", moderngl.TRIANGLES_ADJACENCY),
    ]
    output_types = [
        ("points", moderngl.POINTS),
        ("line_strip", moderngl.LINES),
        ("triangle_strip", moderngl.TRIANGLES),
    ]

    for out_name, out_type in output_types:
        for in_name, in_type in input_types:
            p = ctx.program(
                vertex_shader="""
                #version 330
                void main() {
                    gl_Position = vec4(1.0);
                }
                """,
                geometry_shader=(
                    "#version 330\n"
                    f"layout({in_name}) in;\n"
                    f"layout({out_name}, max_vertices=6) out;\n"
                    "void main() {\n"
                    "    gl_Position = vec4(1.0);\n"
                    "    EmitVertex();\n"
                    "    EndPrimitive();\n"
                    "}\n"
                ),
            )
            assert p.geometry_input == in_type
            assert p.geometry_output == out_type, f"input: {in_name}, output: {out_name}"



import struct
from typing import Any, Dict, List, Tuple


class Attribute:
    def __init__(self):
        self.gl_type = None
        self.program_obj = None
        self.scalar_type = None
        self.rows_length = None
        self.row_length = None
        self.normalizable = None
        self.location = None
        self.array_length = None
        self.dimension = None
        self.shape = None
        self.name = None
        self.extra = None

    def __repr__(self):
        return f"<Attribute: {self.location}>"

    @property
    def mglo(self):
        return self


class Uniform:
    def __init__(self):
        self.program_obj = None
        self.gl_type = None
        self.fmt = None
        self.location = None
        self.array_length = None
        self.element_size = None
        self.dimension = None
        self.name = None
        self.matrix = None
        self.ctx = None
        self.extra = None

    def __repr__(self):
        return f"<Uniform: {self.location}>"

    @property
    def mglo(self):
        return self

    @property
    def value(self):
        data = self.read()
        if self.array_length > 1:
            if self.dimension > 1:
                return [
                    struct.unpack(self.fmt, data[i * self.element_size : i * self.element_size + self.element_size])
                    for i in range(self.array_length)
                ]
            else:
                return [
                    struct.unpack(self.fmt, data[i * self.element_size : i * self.element_size + self.element_size])[0]
                    for i in range(self.array_length)
                ]
        elif self.dimension > 1:
            return struct.unpack(self.fmt, data)
        else:
            return struct.unpack(self.fmt, data)[0]

    @value.setter
    def value(self, value):
        if self.array_length > 1:
            if self.dimension > 1:
                data = b"".join(struct.pack(self.fmt, *row) for row in value)
            else:
                data = b"".join(struct.pack(self.fmt, item) for item in value)
        elif self.dimension > 1:
            data = struct.pack(self.fmt, *value)
        else:
            data = struct.pack(self.fmt, value)
        self.write(data)

    @property
    def handle(self):
        raise NotImplementedError

    @handle.setter
    def handle(self, value):
        return self.ctx._set_uniform_handle(self.program_obj, self.location, value)

    def read(self):
        return self.ctx._read_uniform(
            self.program_obj, self.location, self.gl_type, self.array_length, self.element_size
        )

    def write(self, data: Any):
        self.ctx._write_uniform(self.program_obj, self.location, self.gl_type, self.array_length, data)


class UniformBlock:
    def __init__(self):
        self.program_obj = None
        self.index = None
        self.size = None
        self.name = None
        self.ctx = None
        self.extra = None

    def __repr__(self):
        return f"<UniformBlock: {self.index}>"

    @property
    def mglo(self):
        return self

    @property
    def binding(self):
        return self.ctx._get_ubo_binding(self.program_obj, self.index)

    @binding.setter
    def binding(self, binding):
        self.ctx._set_ubo_binding(self.program_obj, self.index, binding)

    @property
    def value(self):
        return self.ctx._get_ubo_binding(self.program_obj, self.index)

    @value.setter
    def value(self, value):
        self.ctx._set_ubo_binding(self.program_obj, self.index, value)


class StorageBlock:
    def __init__(self):
        self.program_obj = None
        self.index = None
        self.name = None
        self.ctx = None
        self.extra = None

    def __repr__(self):
        return f"<StorageBlock: {self.index}>"

    @property
    def mglo(self):
        return self

    @property
    def binding(self):
        return self.ctx._get_storage_block_binding(self.program_obj, self.index)

    @binding.setter
    def binding(self, binding):
        self.ctx._set_storage_block_binding(self.program_obj, self.index, binding)

    @property
    def value(self):
        return self.ctx._get_storage_block_binding(self.program_obj, self.index)

    @value.setter
    def value(self, value):
        self.ctx._set_storage_block_binding(self.program_obj, self.index, value)


class Subroutine:
    def __init__(self):
        self.index = None
        self.name = None
        self.extra = None

    def __repr__(self):
        return f"<Subroutine: {self.index}>"

    @property
    def mglo(self):
        return self


class Varying:
    def __init__(self):
        self.number = None
        self.array_length = None
        self.dimension = None
        self.name = None
        self.extra = None

    def __repr__(self):
        return f"<Varying: {self.number}>"

    @property
    def mglo(self):
        return self


class Error(Exception):
    pass


ATTRIBUTE_LOOKUP_TABLE = {
    0x1404: (1, 0x1404, 1, 1, False, "i"),
    0x8B53: (2, 0x1404, 1, 2, False, "i"),
    0x8B54: (3, 0x1404, 1, 3, False, "i"),
    0x8B55: (4, 0x1404, 1, 4, False, "i"),
    0x1405: (1, 0x1405, 1, 1, False, "i"),
    0x8DC6: (2, 0x1405, 1, 2, False, "i"),
    0x8DC7: (3, 0x1405, 1, 3, False, "i"),
    0x8DC8: (4, 0x1405, 1, 4, False, "i"),
    0x1406: (1, 0x1406, 1, 1, True, "f"),
    0x8B50: (2, 0x1406, 1, 2, True, "f"),
    0x8B51: (3, 0x1406, 1, 3, True, "f"),
    0x8B52: (4, 0x1406, 1, 4, True, "f"),
    0x140A: (1, 0x140A, 1, 1, False, "d"),
    0x8FFC: (2, 0x140A, 1, 2, False, "d"),
    0x8FFD: (3, 0x140A, 1, 3, False, "d"),
    0x8FFE: (4, 0x140A, 1, 4, False, "d"),
    0x8B5A: (4, 0x1406, 2, 2, True, "f"),
    0x8B65: (6, 0x1406, 2, 3, True, "f"),
    0x8B66: (8, 0x1406, 2, 4, True, "f"),
    0x8B67: (6, 0x1406, 3, 2, True, "f"),
    0x8B5B: (9, 0x1406, 3, 3, True, "f"),
    0x8B68: (12, 0x1406, 3, 4, True, "f"),
    0x8B69: (8, 0x1406, 4, 2, True, "f"),
    0x8B6A: (12, 0x1406, 4, 3, True, "f"),
    0x8B5C: (16, 0x1406, 4, 4, True, "f"),
    0x8F46: (4, 0x140A, 2, 2, False, "d"),
    0x8F49: (6, 0x140A, 2, 3, False, "d"),
    0x8F4A: (8, 0x140A, 2, 4, False, "d"),
    0x8F4B: (6, 0x140A, 3, 2, False, "d"),
    0x8F47: (9, 0x140A, 3, 3, False, "d"),
    0x8F4C: (12, 0x140A, 3, 4, False, "d"),
    0x8F4D: (8, 0x140A, 4, 2, False, "d"),
    0x8F4E: (12, 0x140A, 4, 3, False, "d"),
    0x8F48: (16, 0x140A, 4, 4, False, "d"),
}

UNIFORM_LOOKUP_TABLE = {
    0x8B56: (False, 1, 4, "1i"),
    0x8B57: (False, 2, 8, "2i"),
    0x8B58: (False, 3, 12, "3i"),
    0x8B59: (False, 4, 16, "4i"),
    0x1404: (False, 1, 4, "1i"),
    0x8B53: (False, 2, 8, "2i"),
    0x8B54: (False, 3, 12, "3i"),
    0x8B55: (False, 4, 16, "4i"),
    0x1405: (False, 1, 4, "1I"),
    0x8DC6: (False, 2, 8, "2I"),
    0x8DC7: (False, 3, 12, "3I"),
    0x8DC8: (False, 4, 16, "4I"),
    0x1406: (False, 1, 4, "1f"),
    0x8B50: (False, 2, 8, "2f"),
    0x8B51: (False, 3, 12, "3f"),
    0x8B52: (False, 4, 16, "4f"),
    0x140A: (False, 1, 8, "1d"),
    0x8FFC: (False, 2, 16, "2d"),
    0x8FFD: (False, 3, 24, "3d"),
    0x8FFE: (False, 4, 32, "4d"),
    0x8B5D: (False, 1, 4, "1i"),
    0x8DC0: (False, 1, 4, "1i"),
    0x8DC9: (False, 1, 4, "1i"),
    0x8DCE: (False, 1, 4, "1i"),
    0x8B5E: (False, 1, 4, "1i"),
    0x8DCA: (False, 1, 4, "1i"),
    0x8DD2: (False, 1, 4, "1i"),
    0x8DC1: (False, 1, 4, "1i"),
    0x8DCF: (False, 1, 4, "1i"),
    0x8DD7: (False, 1, 4, "1i"),
    0x8B5F: (False, 1, 4, "1i"),
    0x8DCB: (False, 1, 4, "1i"),
    0x8DD3: (False, 1, 4, "1i"),
    0x8B62: (False, 1, 4, "1i"),
    0x9108: (False, 1, 4, "1i"),
    0x9109: (False, 1, 4, "1i"),
    0x910A: (False, 1, 4, "1i"),
    0x910B: (False, 1, 4, "1i"),
    0x910C: (False, 1, 4, "1i"),
    0x910D: (False, 1, 4, "1i"),
    0x8B60: (False, 1, 4, "1i"),
    0x8DCC: (False, 1, 4, "1i"),
    0x8DD4: (False, 1, 4, "1i"),
    0x904D: (False, 1, 4, "1i"),
    0x8B5A: (True, 4, 16, "4f"),
    0x8B65: (True, 6, 24, "6f"),
    0x8B66: (True, 8, 32, "8f"),
    0x8B67: (True, 6, 24, "6f"),
    0x8B5B: (True, 9, 36, "9f"),
    0x8B68: (True, 12, 48, "12f"),
    0x8B69: (True, 8, 32, "8f"),
    0x8B6A: (True, 12, 48, "12f"),
    0x8B5C: (True, 16, 64, "16f"),
    0x8F46: (True, 4, 32, "4d"),
    0x8F49: (True, 6, 48, "6d"),
    0x8F4A: (True, 8, 64, "8d"),
    0x8F4B: (True, 6, 48, "6d"),
    0x8F47: (True, 9, 72, "9d"),
    0x8F4C: (True, 12, 96, "12d"),
    0x8F4D: (True, 8, 64, "8d"),
    0x8F4E: (True, 12, 96, "12d"),
    0x8F48: (True, 16, 128, "16d"),
}


def make_attribute(name, gl_type, program_obj, location, array_length):
    tmp = ATTRIBUTE_LOOKUP_TABLE.get(gl_type, (1, 0, 1, 1, False, "?"))
    dimension, scalar_type, rows_length, row_length, normalizable, shape = tmp
    rows_length *= array_length
    res = Attribute()
    res.gl_type = gl_type
    res.program_obj = program_obj
    res.scalar_type = scalar_type
    res.rows_length = rows_length
    res.row_length = row_length
    res.normalizable = normalizable
    res.location = location
    res.array_length = array_length
    res.dimension = dimension
    res.shape = shape
    res.name = name
    return res


def make_uniform(name, gl_type, program_obj, location, array_length, ctx):
    tmp = UNIFORM_LOOKUP_TABLE.get(gl_type, (False, 1, 4, "1i"))
    matrix, dimension, element_size, fmt = tmp
    res = Uniform()
    res.name = name
    res.gl_type = gl_type
    res.fmt = fmt
    res.program_obj = program_obj
    res.location = location
    res.array_length = array_length
    res.matrix = matrix
    res.dimension = dimension
    res.element_size = element_size
    res.ctx = ctx
    return res


def make_uniform_block(name, program_obj, index, size, ctx):
    res = UniformBlock()
    res.name = name
    res.program_obj = program_obj
    res.index = index
    res.size = size
    res.ctx = ctx
    return res


def make_storage_block(name, program_obj, index, ctx):
    res = StorageBlock()
    res.name = name
    res.program_obj = program_obj
    res.index = index
    res.ctx = ctx
    return res


def make_subroutine(name, index):
    res = Subroutine()
    res.name = name
    res.index = index
    return res


def make_varying(name, number, array_length, dimension):
    res = Varying()
    res.number = number
    res.name = name
    res.array_length = array_length
    res.dimension = dimension
    return res


class Spv:
    INT32 = 1 << 0
    INT64 = 1 << 1
    UINT32 = 1 << 2
    UINT64 = 1 << 3
    FLOAT32 = 1 << 4
    FLOAT64 = 1 << 5

    VEC2 = 1 << 10
    VEC3 = 1 << 11
    VEC4 = 1 << 12

    MAT2 = 1 << 20
    MAT3 = 1 << 21
    MAT4 = 1 << 22

    NONE = 0
    UNKNOWN = 1 << 30


TRANSLATION_TABLE_SPIRV_GLSL = {
    Spv.INT32: 0x1404,
    Spv.INT32 | Spv.VEC2: 0x8B53,
    Spv.INT32 | Spv.VEC3: 0x8B54,
    Spv.INT32 | Spv.VEC4: 0x8B55,
    Spv.UINT32: 0x1405,
    Spv.UINT32 | Spv.VEC2: 0x8DC6,
    Spv.UINT32 | Spv.VEC3: 0x8DC7,
    Spv.UINT32 | Spv.VEC4: 0x8DC8,
    Spv.FLOAT32: 0x1406,
    Spv.FLOAT32 | Spv.VEC2: 0x8B50,
    Spv.FLOAT32 | Spv.VEC3: 0x8B51,
    Spv.FLOAT32 | Spv.VEC4: 0x8B52,
    Spv.FLOAT64: 0x140A,
    Spv.FLOAT64 | Spv.VEC2: 0x8FFC,
    Spv.FLOAT64 | Spv.VEC3: 0x8FFD,
    Spv.FLOAT64 | Spv.VEC4: 0x8FFE,
    Spv.FLOAT32 | Spv.VEC2 | Spv.MAT2: 0x8B5A,
    Spv.FLOAT32 | Spv.VEC3 | Spv.MAT2: 0x8B65,
    Spv.FLOAT32 | Spv.VEC4 | Spv.MAT2: 0x8B66,
    Spv.FLOAT32 | Spv.VEC2 | Spv.MAT3: 0x8B67,
    Spv.FLOAT32 | Spv.VEC3 | Spv.MAT3: 0x8B5B,
    Spv.FLOAT32 | Spv.VEC4 | Spv.MAT3: 0x8B68,
    Spv.FLOAT32 | Spv.VEC2 | Spv.MAT4: 0x8B69,
    Spv.FLOAT32 | Spv.VEC3 | Spv.MAT4: 0x8B6A,
    Spv.FLOAT32 | Spv.VEC4 | Spv.MAT4: 0x8B5C,
    Spv.FLOAT64 | Spv.VEC2 | Spv.MAT2: 0x8F46,
    Spv.FLOAT64 | Spv.VEC3 | Spv.MAT2: 0x8F49,
    Spv.FLOAT64 | Spv.VEC4 | Spv.MAT2: 0x8F4A,
    Spv.FLOAT64 | Spv.VEC3 | Spv.MAT2: 0x8F4B,
    Spv.FLOAT64 | Spv.VEC3 | Spv.MAT3: 0x8F47,
    Spv.FLOAT64 | Spv.VEC4 | Spv.MAT3: 0x8F4C,
    Spv.FLOAT64 | Spv.VEC2 | Spv.MAT4: 0x8F4D,
    Spv.FLOAT64 | Spv.VEC3 | Spv.MAT4: 0x8F4E,
    Spv.FLOAT64 | Spv.VEC4 | Spv.MAT4: 0x8F48,
}


def parse_spv_inputs(program: int, spv: bytes) -> Dict[int, Attribute]:
    ui32 = struct.Struct("I")
    token = lambda i: ui32.unpack(spv[i * 4 : i * 4 + 4])[0]
    num_tokens = len(spv) // 4

    if token(0) != 0x07230203 or len(spv) % 4 != 0:
        raise ValueError("invalid spv")

    idx = 5

    extracted_names: Dict[int, str] = {}  # id : name
    extracted_storage_class_id: Dict[int, int] = {}  # id : storage_class_id
    extracted_constants: Dict[int, Tuple[int, bytes]] = {}  # const_id : (type_id, content)

    pointer_variable: Dict[int, int] = {}  # id : pointer_type_id
    pointer_allowed_types: Dict[int, int] = {}  # pointer_type_id : type_id
    allowed_types: Dict[int, Tuple[int, int, int, int]] = {}  # type_id : (spv_type, type_id, arr_const_id, arr_length)

    extracted_types: Dict[int, Tuple[int, int]] = {}  # id : (spv_type, arr_length)
    extracted_location: Dict[int, int] = {}  # id : location
    while idx < num_tokens:
        args, opcode = token(idx) >> 16, token(idx) & 0xFFFF
        if opcode == 5:  # OpName
            # We can extract the name of some ids
            name_start, name_end = (idx + 2) * 4, (idx + args) * 4
            extracted_names[token(idx + 1)] = spv[name_start:name_end].rstrip(b"\x00").decode()

        if opcode == 59:  # OpVariable
            # We can extract if it is a vertex shader input or not
            extracted_storage_class_id[token(idx + 2)] = token(idx + 3)

            # Mapping pointer_type_id to id of user variables
            pointer_variable[token(idx + 2)] = token(idx + 1)

        if opcode == 71:  # OpDecorate
            # We can extract the location here
            extracted_location[token(idx + 1)] = token(idx + 3)

        if opcode == 32:  # OpTypePointer
            # Retrieving used (allowed) type_ids for pointer_type_ids
            pointer_allowed_types[token(idx + 1)] = token(idx + 3)

        if opcode == 21:  # OpTypeInt
            unsg, bsz = token(idx + 3), token(idx + 2) // 8
            to_write = Spv.UNKNOWN
            if unsg == 1:
                if bsz == 4:
                    to_write = Spv.UINT32
                elif bsz == 8:
                    to_write = Spv.UINT64
            else:
                if bsz == 4:
                    to_write = Spv.INT32
                elif bsz == 8:
                    to_write = Spv.INT64
            allowed_types[token(idx + 1)] = (to_write, -1, -1, -1)

        if opcode == 22:  # OpTypeFloat
            to_write = Spv.UNKNOWN
            bsz = token(idx + 2) // 8
            if bsz == 4:
                to_write = Spv.FLOAT32
            elif bsz == 8:
                to_write = Spv.FLOAT64
            allowed_types[token(idx + 1)] = (to_write, -1, -1, -1)

        if opcode == 23:  # OpTypeVector
            to_write = Spv.UNKNOWN
            vsz = token(idx + 3)
            if vsz == 2:
                to_write = Spv.VEC2
            elif vsz == 3:
                to_write = Spv.VEC3
            elif vsz == 4:
                to_write = Spv.VEC4

            allowed_types[token(idx + 1)] = (to_write, token(idx + 2), -1, -1)

        if opcode == 24:  # OpTypeMatrix
            to_write = Spv.UNKNOWN
            msz = token(idx + 3)
            if msz == 2:
                to_write = Spv.MAT2
            elif msz == 3:
                to_write = Spv.MAT3
            elif msz == 4:
                to_write = Spv.MAT4

            allowed_types[token(idx + 1)] = (to_write, token(idx + 2), -1, -1)

        if opcode == 28:  # OpTypeArray
            allowed_types[token(idx + 1)] = (Spv.NONE, token(idx + 2), token(idx + 3), -1)

        if opcode == 43:  # OpConstant
            # OpConstant is needed to extract the length of arrays
            # However, it can be used for other purposes
            content_start, content_end = (idx + 3) * 4, (idx + args) * 4
            extracted_constants[token(idx + 2)] = (token(idx + 1), spv[content_start:content_end])

        idx += args

    # Assembly types
    # Some types of variables have pointers to other types of variables used in them.
    # So we use endless research to identify complete types of variables.
    def decode_constant(const_id):
        type_id, content = extracted_constants[const_id]
        typ, _ = assembly(type_id)
        if typ == Spv.INT32:
            return struct.unpack("i", content[:4])[0]
        elif typ == Spv.INT64:
            return struct.unpack("q", content[:8])[0]
        elif typ == Spv.UINT32:
            return struct.unpack("I", content[:4])[0]
        elif typ == Spv.UINT64:
            return struct.unpack("Q", content[:8])[0]
        elif typ == Spv.FLOAT32:
            return struct.unpack("f", content[:4])[0]
        elif typ == Spv.FLOAT64:
            return struct.unpack("d", content[:8])[0]
        else:
            return 0

    def assembly(type_id):
        typ, thrw_typ, arr_const_id, arr_length = allowed_types[type_id]
        if arr_const_id != -1 and arr_length == -1:
            arr_length = decode_constant(arr_const_id)

        if thrw_typ != -1:
            add_typ, multi_arr_length = assembly(thrw_typ)
            typ = add_typ | typ
            arr_length = arr_length * multi_arr_length

        allowed_types[type_id] = (typ, -1, -1, arr_length if arr_length > 0 else 1)

        return typ, arr_length

    for type_id in allowed_types.keys():
        assembly(type_id)
    for ids, pointer_type_id in pointer_variable.items():
        type_id = pointer_allowed_types[pointer_type_id]
        if type_id in allowed_types:
            extracted_types[ids] = (allowed_types[type_id][0], allowed_types[type_id][3])

    # Making a whole list
    exrtacted_general_ids: List[int] = sorted(
        set(
            [
                *extracted_location.keys(),
                *extracted_types.keys(),
                *extracted_storage_class_id.keys(),
                *extracted_names.keys(),
            ]
        )
    )

    extracted_collected: Dict[int, Tuple[str, int, int, int, int]] = {}  # id : variable_info
    for ids in exrtacted_general_ids:
        # to_add: Tuple[str, int, int, int] = ()  # name, class, type, location, arr_length
        name, cls, typ, location, arr_length = "", -1, -1, -1, -1
        if ids in extracted_names:
            name = extracted_names[ids]
        if ids in extracted_storage_class_id:
            cls = extracted_storage_class_id[ids]
        if ids in extracted_types:
            typ, arr_length = extracted_types[ids]
        if ids in extracted_location:
            location = extracted_location[ids]
        extracted_collected[ids] = (name, cls, typ, location, arr_length)

    # Conversion to the GLSL type
    for ids, item in extracted_collected.items():
        if item[2] == -1 or item[2] not in TRANSLATION_TABLE_SPIRV_GLSL:
            continue

        extracted_collected[ids] = (item[0], item[1], TRANSLATION_TABLE_SPIRV_GLSL[item[2]], item[3], item[4])

    # Cropping the data to the required output
    return {
        location: make_attribute(name, gltype, program, location, arr_length)
        for name, cls, gltype, location, arr_length in extracted_collected.values()
        if cls == 1 and location != -1 and gltype != -1
    }


class InvalidObject:
    pass


def resolve_includes(ctx, source):
    def include(match):
        name = match.group(1)
        content = ctx.includes.get(name)
        if content is None:
            raise KeyError(f'cannot include "{name}"')
        return content

    import re

    source = re.sub(r'#include\s+"([^"]+)"', include, source)
    return source




def test_properties(ctx):
    fbo = ctx.simple_framebuffer((4, 4))
    assert fbo.size == (4, 4)
    assert fbo == fbo
    assert fbo.depth_mask is True
    assert fbo.width == 4
    assert fbo.height == 4
    assert fbo.size == (4, 4)
    assert fbo.samples == 0
    assert fbo.glo > 0

    fbo.depth_mask = False
    assert fbo.depth_mask is False


def test_viewport(ctx):
    fbo = ctx.simple_framebuffer((4, 4))
    assert fbo.viewport == (0, 0, 4, 4)
    fbo.viewport = (1, 2, 3, 4)
    assert fbo.viewport == (1, 2, 3, 4)


def test_1(ctx):
    rbo1 = ctx.renderbuffer((4, 4), dtype='f1')
    rbo2 = ctx.renderbuffer((4, 4))
    rbo3 = ctx.renderbuffer((4, 4))
    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer(rbo2)
    fbo3 = ctx.framebuffer(rbo3)
    fbo4 = ctx.framebuffer([rbo1, rbo2, rbo3])

    fbo4.clear(0.0, 1.0, 0.0, 1.0)

    pixels1 = fbo1.read(components=4, dtype='f1')
    assert pixels1 == b'\x00\xff\x00\xff' * 16

    pixels2 = fbo2.read(components=4, dtype='f1')
    assert pixels2 == b'\x00\xff\x00\xff' * 16

    pixels3 = fbo3.read(components=4, dtype='f1')
    assert pixels3 == b'\x00\xff\x00\xff' * 16


def test_2(ctx):
    rbo1 = ctx.renderbuffer((4, 4), dtype='f1')
    rbo2 = ctx.renderbuffer((4, 4))
    rbo3 = ctx.renderbuffer((4, 4))
    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer(rbo2)
    fbo3 = ctx.framebuffer(rbo3)
    fbo4 = ctx.framebuffer([rbo1, rbo2, rbo3])

    fbo1.clear(1.0, 0.0, 0.0, 1.0)
    fbo2.clear(0.0, 1.0, 0.0, 1.0)
    fbo3.clear(0.0, 0.0, 1.0, 1.0)

    pixels1 = fbo4.read(attachment=0, components=4, dtype='f1')
    assert pixels1 == b'\xff\x00\x00\xff' * 16

    pixels2 = fbo4.read(attachment=1, components=4, dtype='f1')
    assert pixels2 == b'\x00\xff\x00\xff' * 16

    pixels3 = fbo4.read(attachment=2, components=4, dtype='f1')
    assert pixels3 == b'\x00\x00\xff\xff' * 16


def test_3(ctx):
    rbo1 = ctx.renderbuffer((4, 4), dtype='f1')
    rbo2 = ctx.renderbuffer((4, 4))
    rbo3 = ctx.renderbuffer((4, 4))
    fbo1 = ctx.framebuffer(rbo1)
    fbo2 = ctx.framebuffer(rbo2)
    fbo3 = ctx.framebuffer(rbo3)
    fbo4 = ctx.framebuffer([rbo1, rbo2, rbo3])

    fbo1.clear(1.0, 0.0, 0.0, 1.0)
    fbo2.clear(0.0, 1.0, 0.0, 1.0)
    fbo3.clear(0.0, 0.0, 1.0, 1.0)

    pixels1 = bytearray(64)
    fbo4.read_into(pixels1, attachment=0, components=4, dtype='f1')
    assert bytes(pixels1) == b'\xff\x00\x00\xff' * 16

    pixels2 = bytearray(64)
    fbo4.read_into(pixels2, attachment=1, components=4, dtype='f1')
    assert bytes(pixels2) == b'\x00\xff\x00\xff' * 16

    pixels3 = bytearray(64)
    fbo4.read_into(pixels3, attachment=2, components=4, dtype='f1')
    assert bytes(pixels3) == b'\x00\x00\xff\xff' * 16
